# -*- coding: utf-8 -*-
"""
Known Interactions that use AbstractInteraction:
    pt.MatchInteraction2
    pt.MultiImageInteraction
    wbia.NameInteraction
"""
import six
import re
import utool as ut
import matplotlib as mpl

ut.noinject(__name__, '[abstract_iteract]')
from . import draw_func2 as df2  # NOQA
from wbia.plottool import fig_presenter  # NOQA
from wbia.plottool import plot_helpers as ph  # NOQA
from wbia.plottool import interact_helpers as ih  # NOQA

# (print, print_, printDBG, rrr, profile) = utool.inject(__name__,
# '[abstract_iteract]')

DEBUG = ut.get_argflag('--debug-interact')
VERBOSE = ut.VERBOSE or True


# for scoping
__REGISTERED_INTERACTIONS__ = []


def register_interaction(self):
    global __REGISTERED_INTERACTIONS__
    if VERBOSE:
        print('[pt] Registering intearction: self=%r' % (self,))
    __REGISTERED_INTERACTIONS__.append(self)
    if VERBOSE:
        print(
            '[pt] There are now %d registered interactions'
            % (len(__REGISTERED_INTERACTIONS__))
        )


def unregister_interaction(self):
    global __REGISTERED_INTERACTIONS__
    if VERBOSE:
        print('[pt] Unregistering intearction: self=%r' % (self,))
    try:
        __REGISTERED_INTERACTIONS__.remove(self)
    except ValueError:
        pass
    if VERBOSE:
        print(
            '[pt] There are now %d registered interactions'
            % (len(__REGISTERED_INTERACTIONS__))
        )


class AbstractInteraction(object):
    """
    An interaction is meant to take up an entire figure

    overwrite either self.plot(fnum, pnum) or self.staic_plot(fnum, pnum) or show_page
    """

    LEFT_BUTTON = 1
    MIDDLE_BUTTON = 2
    RIGHT_BUTTON = 3

    MOUSE_BUTTONS = {
        LEFT_BUTTON: 'left',
        MIDDLE_BUTTON: 'middle',
        RIGHT_BUTTON: 'right',
    }

    def __init__(self, **kwargs):
        debug = kwargs.get('debug', None)
        self.debug = DEBUG if debug is None else debug
        if self.debug:
            print('[pt.a] create interaction')
        self.fnum = kwargs.get('fnum', None)
        if self.fnum is None:
            self.fnum = df2.next_fnum()
        self.interaction_name = kwargs.get('interaction_name', 'AbstractInteraction')
        # Careful this might cause memory leaks
        self.scope = []  # for keeping those widgets alive!
        self.is_down = {}
        self.is_drag = {}
        self.is_running = False

        self.pan_event_list = []
        self.zoom_event_list = []

        self.fig = getattr(self, 'fig', None)

        for button in self.MOUSE_BUTTONS.values():
            self.is_down[button] = None
            self.is_drag[button] = None

        autostart = kwargs.get('autostart', False)
        if autostart:
            self.start()

    def reset_mouse_state(self):
        for key in self.is_down.keys():
            self.is_down[key] = False
        for key in self.is_drag.keys():
            self.is_drag[key] = False

    def enable_pan_and_zoom(self, ax):
        self.enable_zoom(ax)
        self.enable_pan(ax)

    def enable_pan(self, ax):
        from wbia.plottool.interactions import PanEvents

        pan = PanEvents(ax)
        self.pan_event_list.append(pan)

    def enable_zoom(self, ax):
        from wbia.plottool.interactions import zoom_factory

        self.zoom_event_list.append(zoom_factory(ax))

    def _start_interaction(self):
        # self.fig = df2.figure(fnum=self.fnum, doclf=True, docla=True)
        self.fig = df2.figure(fnum=self.fnum, doclf=True)
        ih.connect_callback(self.fig, 'close_event', self.on_close)
        register_interaction(self)
        self.is_running = True

    def _ensure_running(self):
        if not self.is_running:
            self._start_interaction()

    def start(self):
        self._ensure_running()
        self.show_page()
        self.show()

    def print_status(self):
        print('is_down = ' + ut.repr2(self.is_down))
        print('is_drag = ' + ut.repr2(self.is_drag))

    def _preshow_page(self):
        self._ensure_running()
        if self.debug:
            print('[pt.a] show page')
        self.fig = ih.begin_interaction(self.interaction_name, fnum=self.fnum)

    def _postshow_page(self):
        self.connect_callbacks()

    def _show_page(self):
        if hasattr(self, 'plot'):
            self.plot(fnum=self.fnum, pnum=(1, 1, 1))
        else:
            self.static_plot(fnum=self.fnum, pnum=(1, 1, 1))

    def show_page(self, *args):
        """
        Hack: this function should probably not be defined, but it is for
        convinience of a developer.
        Override this or create static plot function
        (preferably override)
        """
        self._preshow_page()
        self._show_page()
        self._postshow_page()

    def connect_callbacks(self):
        if self.debug:
            print('[pt.a] connect_callbacks')
        ih.connect_callback(self.fig, 'button_press_event', self.on_click)
        ih.connect_callback(self.fig, 'button_release_event', self.on_click_release)
        ih.connect_callback(self.fig, 'key_press_event', self.on_key_press)
        ih.connect_callback(self.fig, 'motion_notify_event', self.on_motion)
        ih.connect_callback(self.fig, 'draw_event', self.on_draw)
        ih.connect_callback(self.fig, 'scroll_event', self.on_scroll)

    def bring_to_front(self):
        import utool

        with utool.embed_on_exception_context:
            fig_presenter.bring_to_front(self.fig)

    def draw(self):
        if self.debug > 5:
            print('[pt.a] draw')
        self.fig.canvas.draw()

    def on_draw(self, event=None):
        if self.debug > 5:
            print('[pt.a] on draw')
        pass

    def show(self):
        if self.debug:
            print('[pt.a] show')
        self.fig.show()

    def update(self):
        if self.debug:
            print('[pt.a] update')
        # fig_presenter.update()
        self.fig.canvas.update()
        self.fig.canvas.flush_events()

    def on_scroll(self, event):
        if self.debug:
            print('[pt.a] on_scroll')
            print(ut.repr3(event.__dict__))
        pass

    def close(self):
        assert isinstance(self.fig, mpl.figure.Figure)
        fig_presenter.close_figure(self.fig)

    def on_close(self, event=None):
        print('[pt] handling interaction close')
        unregister_interaction(self)
        self.is_running = False

    def on_motion(self, event):
        if self.debug > 5:
            print('[pt.a] on_motion')
        for button in self.MOUSE_BUTTONS.values():
            if self.is_down[button]:
                if not self.is_drag[button]:
                    self.is_drag[button] = True
                    self.on_drag_start(event)
        if any(self.is_drag.values()):
            self.on_drag(event)

        for pan in self.pan_event_list:
            pan.pan_on_motion(event)
        # print('event = %r' % (event.__dict__,))
        pass

    def on_drag(self, event=None):
        if self.debug > 1:
            print('[pt.a] on_drag')
        if ih.clicked_inside_axis(event):
            self.on_drag_inside(event)
        # Make sure BLIT (bit block transfer) is used for updates
        # self.fig.canvas.blit(self.fig.ax.bbox)
        pass

    def on_drag_inside(self, event=None):
        if self.debug > 1:
            print('[pt.a] on_drag_inside')

    def on_drag_stop(self, event=None):
        if self.debug > 0:
            print('[pt.a] on_drag_stop')
        pass

    def on_drag_start(self, event=None):
        if self.debug > 0:
            print('[pt.a] on_drag_start')
        pass

    def on_key_press(self, event):
        if self.debug > 0:
            print('[pt.a] on_key_press')
        # self.print_status()
        pass

    def on_click(self, event):
        if self.debug > 0:
            print('[pt.a] on_click')
            # print('[pt.a] on_click. event=%r' % (ut.repr2(event.__dict__)))
        # raise NotImplementedError('implement yourself')
        if event.button is not None:
            for button in self.MOUSE_BUTTONS.values():
                if self.MOUSE_BUTTONS[event.button] == button:
                    self.is_down[button] = True
        # if event.button == self.LEFT_BUTTON:
        #    self.is_down['left'] = True
        # if event.button == self.RIGHT_BUTTON:
        #    self.is_down['right'] = True
        # if event.button == self.MIDDLE_BUTTON:
        #    self.is_down['middle'] = True
        if ih.clicked_inside_axis(event):
            ax = event.inaxes
            self.on_click_inside(event, ax)
        else:
            self.on_click_outside(event)

        for pan in self.pan_event_list:
            pan.pan_on_press(event)

    def on_click_release(self, event):
        if self.debug > 0:
            print('[pt.a] on_release')
        for button in self.MOUSE_BUTTONS.values():
            flag = (
                event is None
                or event.button is None
                or self.MOUSE_BUTTONS[event.button] == button
            )
            if flag:
                self.is_down[button] = False
                if self.is_drag[button]:
                    self.is_drag[button] = False
                    self.on_drag_stop(event)
        # if event.button == self.LEFT_BUTTON:
        #    self.is_down['left'] = False
        #    self.is_drag['left'] = False
        # if event.button == self.RIGHT_BUTTON:
        #    self.is_down['right'] = False
        # if event.button == self.MIDDLE_BUTTON:
        #    self.is_down['middle'] = False
        for pan in self.pan_event_list:
            pan.pan_on_release(event)

    def on_click_inside(self, event, ax):
        pass

    def on_click_outside(self, event):
        pass

    def show_popup_menu(self, options, event):
        """
        context menu
        """
        import wbia.guitool as gt

        height = self.fig.canvas.geometry().height()
        qpoint = gt.newQPoint(event.x, height - event.y)
        qwin = self.fig.canvas
        gt.popup_menu(qwin, qpoint, options)

    def clear_parent_axes(self, ax):
        """ for clearing axes that we appended anything to """
        child_axes = ph.get_plotdat(ax, 'child_axes', [])
        ph.set_plotdat(ax, 'child_axes', [])
        for subax in child_axes:
            to_remove = None
            for tup in self.scope:
                if tup[1] is subax:
                    to_remove = tup
                    break
            if to_remove is not None:
                self.scope.remove(to_remove)
            subax.cla()
            self.fig.delaxes(subax)
        ph.del_plotdat(ax, df2.DF2_DIVIDER_KEY)
        ax.cla()

    # def make_hud(self):

    # def prepare_page(self, pagenum):
    #    self.clean_scope()

    def clean_scope(self):
        """ Removes any widgets saved in the interaction scope """
        self.scope = []

    def append_button(
        self,
        text,
        divider=None,
        rect=None,
        callback=None,
        size='9%',
        location='bottom',
        ax=None,
        **kwargs,
    ):
        """ Adds a button to the current page """
        if rect is not None:
            new_ax = df2.plt.axes(rect)
        if rect is None and divider is None:
            if ax is None:
                ax = df2.gca()
            divider = df2.ensure_divider(ax)
        if divider is not None:
            new_ax = divider.append_axes(location, size=size, pad=0.05)
        if callback is not None:
            color, hovercolor = '.85', '.95'
        else:
            color, hovercolor = '.88', '.88'
            # color, hovercolor = u'.45', u'.45'
        # if isinstance(text, six.text_type):
        new_but = mpl.widgets.Button(new_ax, text, color=color, hovercolor=hovercolor)
        # elif isinstance(text, (list, tuple)):
        #    labels = [False] * len(text)
        #    labels[0] = True
        #    new_but = mpl.widgets.CheckButtons(new_ax, text, labels)
        # else:
        #    raise ValueError('bad input')

        if callback is not None:
            new_but.on_clicked(callback)
        else:
            button_text = new_but.ax.texts[0]
            button_text.set_color('.6')
            # button_text.set_color('r')
            # ut.embed()
            # print('new_but.color = %r' % (new_but.color,))
        # else:
        # TODO: figure ou how to gray out these buttons
        #    new_but.color = u'.1'
        #    new_but.hovercolor = u'.1'
        #    new_but.active = False
        #    print('new_but.color = %r' % (new_but.color,))
        ph.set_plotdat(new_ax, 'viztype', 'button')
        ph.set_plotdat(new_ax, 'text', text)
        # ph.set_plotdat(new_ax, 'parent_axes', ax)
        if ax is not None:
            child_axes = ph.get_plotdat(ax, 'child_axes', [])
            child_axes.append(new_ax)
            ph.set_plotdat(ax, 'child_axes', child_axes)
        for key, val in six.iteritems(kwargs):
            ph.set_plotdat(new_ax, key, val)
        # Keep buttons from losing scrop
        tup = (new_but, new_ax)
        self.scope.append(tup)
        return tup


class AbstractPagedInteraction(AbstractInteraction):
    def __init__(self, nPages=None, draw_hud=True, **kwargs):
        self.current_pagenum = 0
        assert nPages is not None
        self.nPages = nPages
        self.draw_hud = draw_hud
        self.NEXT_PAGE_HOTKEYS = ['right', 'pagedown']
        self.PREV_PAGE_HOTKEYS = ['left', 'pageup']
        super(AbstractPagedInteraction, self).__init__(**kwargs)

    def next_page(self, event):
        # print('next')
        self.show_page(self.current_pagenum + 1)
        pass

    def prev_page(self, event):
        if self.current_pagenum == 0:
            return
        # print('prev')
        self.show_page(self.current_pagenum - 1)
        pass

    def make_hud(self):
        """ Creates heads up display """
        import wbia.plottool as pt

        if not self.draw_hud:
            return
        # Button positioning
        # w, h = .08, .04
        # w, h = .14, .08
        w, h = 0.14, 0.07
        hl_slot, hr_slot = pt.make_bbox_positioners(
            y=0.02, w=w, h=h, xpad=0.05, startx=0, stopx=1
        )
        prev_rect = hl_slot(0)
        next_rect = hr_slot(0)
        # print('prev_rect = %r' % (prev_rect,))
        # print('next_rect = %r' % (next_rect,))

        # Create buttons
        prev_callback = None if self.current_pagenum == 0 else self.prev_page
        next_callback = (
            None if self.current_pagenum == self.nPages - 1 else self.next_page
        )
        prev_text = 'prev\n' + pretty_hotkey_map(self.PREV_PAGE_HOTKEYS)
        next_text = 'next\n' + pretty_hotkey_map(self.NEXT_PAGE_HOTKEYS)
        self.append_button(prev_text, callback=prev_callback, rect=prev_rect)
        self.append_button(next_text, callback=next_callback, rect=next_rect)

    def prepare_page(self, fulldraw=True):
        import wbia.plottool as pt

        ih.disconnect_callback(self.fig, 'button_press_event')
        ih.disconnect_callback(self.fig, 'button_release_event')
        ih.disconnect_callback(self.fig, 'key_press_event')
        ih.disconnect_callback(self.fig, 'motion_notify_event')

        figkw = {
            'fnum': self.fnum,
            'doclf': fulldraw,
            'docla': fulldraw,
        }
        if fulldraw:
            self.fig = pt.figure(**figkw)
        self.make_hud()
        ih.connect_callback(self.fig, 'button_press_event', self.on_click)
        ih.connect_callback(self.fig, 'button_release_event', self.on_click_release)
        ih.connect_callback(self.fig, 'key_press_event', self.on_key_press)
        ih.connect_callback(self.fig, 'motion_notify_event', self.on_motion)

    def on_key_press(self, event):
        if matches_hotkey(event.key, self.PREV_PAGE_HOTKEYS):
            if self.current_pagenum != 0:
                self.prev_page(event)
        if matches_hotkey(event.key, self.NEXT_PAGE_HOTKEYS):
            if self.current_pagenum != self.nPages - 1:
                self.next_page(event)
        self.draw()


# moved to interactions.py
# def zoom_factory(ax, zoomable_list, base_scale=1.1):
#    """
#    TODO: make into interaction

#    References:
#        https://gist.github.com/tacaswell/3144287
#    """
#    def zoom_fun(event):
#        #print('zooming')
#        # get the current x and y limits
#        cur_xlim = ax.get_xlim()
#        cur_ylim = ax.get_ylim()
#        xdata = event.xdata  # get event x location
#        ydata = event.ydata  # get event y location
#        if xdata is None or ydata is None:
#            return
#        if event.button == 'up':
#            # deal with zoom in
#            scale_factor = 1 / base_scale
#        elif event.button == 'down':
#            # deal with zoom out
#            scale_factor = base_scale
#        else:
#            raise NotImplementedError('event.button=%r' % (event.button,))
#            # deal with something that should never happen
#            scale_factor = 1
#            print(event.button)
#        for zoomable in zoomable_list:
#            zoom = zoomable.get_zoom()
#            new_zoom = zoom / (scale_factor ** (1.2))
#            zoomable.set_zoom(new_zoom)
#        # Get distance from the cursor to the edge of the figure frame
#        x_left = xdata - cur_xlim[0]
#        x_right = cur_xlim[1] - xdata
#        y_top = ydata - cur_ylim[0]
#        y_bottom = cur_ylim[1] - ydata
#        ax.set_xlim([xdata - x_left * scale_factor, xdata + x_right * scale_factor])
#        ax.set_ylim([ydata - y_top * scale_factor, ydata + y_bottom * scale_factor])

#        # ----
#        ax.figure.canvas.draw()  # force re-draw

#    fig = ax.get_figure()  # get the figure of interest
#    # attach the call back
#    fig.canvas.mpl_connect('scroll_event', zoom_fun)

#    #return the function
#    return zoom_fun


def pretty_hotkey_map(hotkeys):
    if hotkeys is None:
        return ''
    hotkeys = [hotkeys] if not isinstance(hotkeys, list) else hotkeys
    mapping = {
        # 'right': 'right arrow',
        # 'left':  'left arrow',
    }
    mapped_hotkeys = [mapping.get(hk, hk) for hk in hotkeys]
    hotkey_str = '(' + ut.conj_phrase(mapped_hotkeys, 'or') + ')'
    return hotkey_str


def matches_hotkey(key, hotkeys):
    hotkeys = [hotkeys] if not isinstance(hotkeys, list) else hotkeys
    # flags = [re.match(hk, '^' + key + '$') for hk in hotkeys]
    flags = [re.match(hk, key) is not None for hk in hotkeys]
    return any(flags)



# -*- coding: utf-8 -*-
"""
DEPRICATE FOR CORE ANNOT AND CORE IMAGE DEFS
"""
import logging
import utool as ut
import six
import copy

# from wbia import dtool
from os.path import join
from os.path import splitext
from six.moves import zip, map, range, filter  # NOQA
from wbia import constants as const
from utool._internal.meta_util_six import get_funcname

(print, rrr, profile) = ut.inject2(__name__, '[cfg]')
logger = logging.getLogger('wbia')

# ConfigBase = ut.DynStruct
# ConfigBase = object
ConfigBase = ut.Pref


def parse_config_items(cfg):
    """
    Recursively extracts key, val pairs from Config objects
    into a flat list. (there must not be name conflicts)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.Config import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> cfg = ibs.cfg.query_cfg
        >>> param_list = parse_config_items(cfg)
    """
    # import wbia
    param_list = []
    seen = set([])
    for item in cfg.items():
        key, val = item
        # if isinstance(val, wbia.algo.Config.ConfigBase):
        if isinstance(val, ConfigBase):
            child_cfg = val
            param_list.extend(parse_config_items(child_cfg))
            # logger.info(key)
            pass
        elif key.startswith('_'):
            # logger.info(key)
            pass
        else:
            if key in seen:
                logger.info('[Config] WARNING: key=%r appears more than once' % (key,))
            seen.add(key)
            param_list.append(item)
            # logger.info(key)
    return param_list


def make_config_metaclass():
    """
    Creates a metaclass for Config objects that automates some of the more
    tedious functions to write

    Like:
        get_cfgstr
        and the comparison methods


    Example:
        from wbia.algo.Config import *  # NOQA
        @six.add_metaclass(ConfigMetaclass)
        class FooConfig(ConfigBase):
            def __init__(cfg):
                super(FooConfig, cfg).__init__(name='FooConfig')
                cfg.initialize_params()

            def get_param_info_list(cfg):
                return [
                    ut.ParamInfo('x', 'y'),
                    ut.ParamInfo('z', 3),
                ]
        cfg = FooConfig()
        logger.info(cfg.get_cfgstr(ignore_keys=['x']))
        logger.info(cfg.get_cfgstr(ignore_keys=[]))

        cfg = GenericConfig()
        cfg.x = 'y'

    """
    methods_list = ut.get_comparison_methods()

    # Decorator for functions that we will inject into our metaclass
    def _register(func):
        methods_list.append(func)
        return func

    @_register
    def get_cfgstr_list(cfg, ignore_keys=None, **kwargs):
        """ default get_cfgstr_list, can be overrided by a config object """
        if hasattr(cfg, 'get_param_info_list'):
            if ignore_keys is not None:
                itemstr_list = [
                    pi.get_itemstr(cfg)
                    for pi in cfg.get_param_info_list()
                    if pi.varname not in ignore_keys
                ]
            else:
                itemstr_list = [pi.get_itemstr(cfg) for pi in cfg.get_param_info_list()]
        else:
            try:
                item_list = parse_config_items(cfg)
                assert item_list is not None
                if ignore_keys is None:
                    itemstr_list = [
                        key + '=' + six.text_type(val) for key, val in item_list
                    ]
                else:
                    itemstr_list = [
                        key + '=' + six.text_type(val)
                        for key, val in item_list
                        if key not in ignore_keys
                    ]
            except Exception as ex:
                logger.info(ignore_keys is None)
                logger.info(ignore_keys)
                ut.printex(ex, keys=['item_list', 'ignore_keys'])
                raise
        filtered_itemstr_list = list(filter(len, itemstr_list))
        config_name = cfg.get_config_name()
        body = ','.join(filtered_itemstr_list)
        cfgstr = ''.join(['_', config_name, '(', body, ')'])
        return cfgstr

    @_register
    @profile
    def initialize_params(cfg):
        """ Initializes config class attributes based on params info list """
        for pi in cfg.get_param_info_list():
            setattr(cfg, pi.varname, pi.default)

    @_register
    def parse_items(cfg, **kwargs):
        return parse_config_items(cfg)

    @_register
    def keys(cfg, **kwargs):
        return ut.take_column(cfg.parse_items(), 0)

    @_register
    def get_config_name(cfg, **kwargs):
        """ the user might want to overwrite this function """
        class_str = six.text_type(cfg.__class__)
        full_class_str = class_str.replace("<class '", '').replace("'>", '')
        config_name = splitext(full_class_str)[1][1:].replace('Config', '')
        return config_name

    @_register
    def __hash__(cfg):
        """ Needed for comparison operators """
        return hash(cfg.get_cfgstr())

    @_register
    def get_cfgstr(cfg, **kwargs):
        return ''.join(cfg.get_cfgstr_list(**kwargs))

    @_register
    def lookup_paraminfo(cfg, key):
        for pi in cfg.get_param_info_list():
            if pi.varname == key:
                return pi
        raise KeyError('no such param info (in the old config)')

    class ConfigMetaclass(type):
        """
        Defines extra methods for Configs
        """

        def __new__(cls, name, bases, dct):
            """
            cls - meta
            name - classname
            supers - bases
            dct - class dictionary
            """
            # assert 'get_cfgstr_list' in dct, (
            #  'must have defined get_cfgstr_list.  name=%r' % (name,))
            # Inject registered function
            for func in methods_list:
                if get_funcname(func) not in dct:
                    funcname = get_funcname(func)
                    dct[funcname] = func
                else:
                    funcname = get_funcname(func)
                    dct['meta_' + funcname] = func
                # ut.inject_func_as_method(metaself, func)
            return type.__new__(cls, name, bases, dct)

    return ConfigMetaclass


ConfigMetaclass = make_config_metaclass()


@six.add_metaclass(ConfigMetaclass)
class GenericConfig(ConfigBase):
    def __init__(cfg, *args, **kwargs):
        super(GenericConfig, cfg).__init__(*args, **kwargs)


@six.add_metaclass(ConfigMetaclass)
class NNConfig(ConfigBase):
    r"""
    CommandLine:
        python -m wbia.algo.Config --exec-NNConfig

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.Config import *  # NOQA
        >>> nn_cfg = NNConfig()
        >>> nn_cfg = NNConfig(requery=True)
        >>> result = nn_cfg.get_cfgstr()
        >>> print(result)
        _NN(single,K=4,Kn=1,padk=False,cks800)
    """

    def __init__(nn_cfg, **kwargs):
        super(NNConfig, nn_cfg).__init__()
        # if True:
        nn_cfg.initialize_params()
        # else:
        #    nn_cfg.K = 4
        #    # TODO: force to false when in vsone
        #    nn_cfg.use_k_padding = False
        #    # number of annots before a new multi-indexer is built
        #    #nn_cfg.min_reindex_thresh = 3
        #    #nn_cfg.index_method = 'multi'
        #    nn_cfg.index_method = 'single'
        #    nn_cfg.Knorm = 1
        #    nn_cfg.checks = 800
        # number of annots before a new multi-indexer is built
        nn_cfg.min_reindex_thresh = 200
        # number of annots before a new multi-indexer is built
        # nn_cfg.max_subindexers = 2
        # nn_cfg.valid_index_methods = ['single', 'multi', 'name']
        nn_cfg.valid_index_methods = ['single']
        nn_cfg.update(**kwargs)

    def make_feasible(nn_cfg):
        # normalizer rule depends on Knorm
        assert nn_cfg.index_method in nn_cfg.valid_index_methods

    def get_param_info_list(nn_cfg):
        # new way to try and specify config options.
        # not sure if i like it yet
        param_info_list = ut.flatten(
            [
                [
                    ut.ParamInfo('index_method', 'single', ''),
                    ut.ParamInfo('K', 4, type_=int),
                    ut.ParamInfo('Knorm', 1, 'Kn='),
                    ut.ParamInfo('use_k_padding', False, 'padk='),
                    ut.ParamInfo('requery', False, type_=bool, hideif=False),
                    # ut.ParamInfo('condrecover', True, type_=bool, hideif=False),
                    ut.ParamInfo('checks', 800, 'cks', type_=int),
                    # ut.ParamInfo('ratio_thresh', None, type_=float, hideif=None),
                ],
            ]
        )
        return param_info_list


@six.add_metaclass(ConfigMetaclass)
class SpatialVerifyConfig(ConfigBase):
    """
    Spatial verification
    """

    def __init__(sv_cfg, **kwargs):
        super(SpatialVerifyConfig, sv_cfg).__init__(name='sv_cfg')
        tau = 6.28  # 318530
        sv_cfg.sv_on = True
        sv_cfg.xy_thresh = 0.01
        sv_cfg.scale_thresh = 2.0
        sv_cfg.ori_thresh = tau / 4.0
        sv_cfg.min_nInliers = 4
        sv_cfg.full_homog_checks = True
        # sv_cfg.nNameShortlistSVER = 50
        sv_cfg.nNameShortlistSVER = 40
        # sv_cfg.nAnnotPerNameSVER = 6
        sv_cfg.nAnnotPerNameSVER = 3
        # sv_cfg.prescore_method = 'csum'
        sv_cfg.prescore_method = 'nsum'
        sv_cfg.use_chip_extent = True  # BAD CONFIG?
        # weight feature scores with sver errors
        sv_cfg.sver_output_weighting = False
        sv_cfg.refine_method = 'homog'
        # weight feature scores with sver errors
        sv_cfg.weight_inliers = True
        sv_cfg.update(**kwargs)

    def get_cfgstr_list(sv_cfg, **kwargs):
        if not sv_cfg.sv_on or sv_cfg.xy_thresh is None:
            return ['_SV(OFF)']
        thresh_tup = (sv_cfg.xy_thresh, sv_cfg.scale_thresh, sv_cfg.ori_thresh)
        thresh_str = ut.remove_chars(six.text_type(thresh_tup), ' ()').replace(',', ';')
        sv_cfgstr = [
            '_SV(',
            thresh_str,
            'minIn=%d,' % (sv_cfg.min_nInliers,),
            'nNRR=%d,' % (sv_cfg.nNameShortlistSVER,),
            'nARR=%d,' % (sv_cfg.nAnnotPerNameSVER,),
            sv_cfg.prescore_method,
            ',',
            'cdl,' * sv_cfg.use_chip_extent,  # chip diag len
            '+ow,' * sv_cfg.sver_output_weighting,  # chip diag len
            '+wi,' * sv_cfg.weight_inliers,  # chip diag len
            '+fc,' * sv_cfg.full_homog_checks,
        ]

        if sv_cfg.refine_method != 'homog':
            sv_cfgstr += [sv_cfg.refine_method]

        sv_cfgstr += [
            ')',
        ]
        return sv_cfgstr


@six.add_metaclass(ConfigMetaclass)
class AggregateConfig(ConfigBase):
    """
    Old Agg Cfg
    """

    def __init__(agg_cfg, **kwargs):
        super(AggregateConfig, agg_cfg).__init__(name='agg_cfg')
        # chipsum, namesum,
        # agg_cfg.score_method = 'csum'
        agg_cfg.score_method = 'nsum'
        alt_methods = {
            'topk': 'topk',
            'chipsum': 'csum',
            'namesum': 'nsum',
        }
        # For Placket-Luce
        agg_cfg.max_alts = 50
        # -----
        # User update
        agg_cfg.update(**kwargs)
        # ---
        key = agg_cfg.score_method.lower()
        # Use w as a toggle for weighted mode
        # Sanatize the scoring method
        if key in alt_methods:
            agg_cfg.score_method = alt_methods[key]

    def get_cfgstr_list(agg_cfg, **kwargs):
        agg_cfgstr = []
        agg_cfgstr.append('_AGG(')
        agg_cfgstr.append(agg_cfg.score_method)
        agg_cfgstr.append(')')
        return agg_cfgstr


@six.add_metaclass(ConfigMetaclass)
class FlannConfig(ConfigBase):
    r"""
    this flann is only for neareset neighbors in vsone/many
    TODO: this might not need to be here, should be part of neighbor config

    References:
        http://www.cs.ubc.ca/research/flann/uploads/FLANN/flann_pami2014.pdf
        http://www.cs.ubc.ca/research/flann/uploads/FLANN/flann_manual-1.8.4.pdf
        http://docs.opencv.org/trunk/modules/flann/doc/flann_fast_approximate_nearest_neighbor_search.html
    """

    def __init__(flann_cfg, **kwargs):
        super(FlannConfig, flann_cfg).__init__(name='flann_cfg')
        # General Params
        flann_cfg.algorithm = 'kdtree'  # linear
        flann_cfg.flann_cores = 0  # doesnt change config, just speed
        # KDTree params
        flann_cfg.trees = 8
        # KMeansTree params
        flann_cfg.iterations = 11
        flann_cfg.centers_init = 'random'
        flann_cfg.cb_index = 0.4
        flann_cfg.branching = 64
        # THESE CONFIGS DONT BELONG TO FLANN. THEY ARE INDEXER CONFIGS
        flann_cfg.fgw_thresh = None
        flann_cfg.minscale_thresh = None
        flann_cfg.maxscale_thresh = None
        flann_cfg.update(**kwargs)

    def get_flann_params(flann_cfg):
        flann_params = dict(
            algorithm=flann_cfg.algorithm,
            trees=flann_cfg.trees,
            cores=flann_cfg.flann_cores,
        )
        return flann_params

    def get_cfgstr_list(flann_cfg, **kwargs):
        flann_cfgstrs = ['_FLANN(']
        if flann_cfg.algorithm == 'kdtree':
            flann_cfgstrs += ['%d_kdtrees' % flann_cfg.trees]
        elif flann_cfg.algorithm == 'kdtree':
            flann_cfgstrs += [
                '%s_' % flann_cfg.algorithm,
                'iter=%s_' % flann_cfg.iterations,
                'cb=%s_' % flann_cfg.cb_index,
                'branch=%s' % flann_cfg.branching,
            ]
        elif flann_cfg.algorithm == 'linear':
            flann_cfgstrs += ['%s' % flann_cfg.algorithm]
        else:
            flann_cfgstrs += ['%s' % flann_cfg.algorithm]
        if flann_cfg.fgw_thresh is not None and flann_cfg.fgw_thresh > 0:
            # HACK FOR GGR
            flann_cfgstrs += ['_fgwthrsh=%s' % flann_cfg.fgw_thresh]
        if (flann_cfg.minscale_thresh is not None) or (
            flann_cfg.maxscale_thresh is not None
        ):
            # HACK FOR GGR
            flann_cfgstrs += [
                'scalethrsh=%s,%s'
                % (flann_cfg.minscale_thresh, flann_cfg.maxscale_thresh)
            ]
        # flann_cfgstrs += ['checks=%r' % flann_cfg.checks]
        flann_cfgstrs += [')']
        return flann_cfgstrs


@six.add_metaclass(ConfigMetaclass)
class NNWeightConfig(ConfigBase):
    r"""
    CommandLine:
        python -m wbia.algo.Config --test-NNWeightConfig

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.Config import *  # NOQA
        >>> cfg_list = [
        ...     NNWeightConfig(),
        ...     NNWeightConfig(can_match_sameimg=True, can_match_samename=False),
        ...     NNWeightConfig(ratio_thresh=.625, lnbnn_on=False),
        ...     NNWeightConfig(ratio_thresh=.625, lnbnn_normer='foobarstr'),
        ... ]
        >>> result = '\n'.join([cfg.get_cfgstr() for cfg in cfg_list])
        >>> print(result)
        _NNWeight(lnbnn,fg,last,nosqrd_dist)
        _NNWeight(lnbnn,fg,last,sameimg,nosamename,nosqrd_dist)
        _NNWeight(ratio_thresh=0.625,fg,last,nosqrd_dist)
        _NNWeight(ratio_thresh=0.625,lnbnn,fg,last,lnbnn_normer=foobarstr,lnbnn_norm_thresh=0.5,nosqrd_dist)
    """

    @profile
    def __init__(nnweight_cfg, **kwargs):
        super(NNWeightConfig, nnweight_cfg).__init__(name='nnweight_cfg')
        nnweight_cfg.initialize_params()
        nnweight_cfg.update(**kwargs)

    @profile
    def get_param_info_list(nnweight_cfg):
        # new way to try and specify config options.
        # not sure if i like it yet
        param_info_list = ut.flatten(
            [
                [
                    ut.ParamInfo('ratio_thresh', None, type_=float, hideif=None),
                    ut.ParamInfoBool('lnbnn_on', True, hideif=False),
                    ut.ParamInfoBool('const_on', False, hideif=False),
                    ut.ParamInfoBool('lograt_on', False, hideif=False),
                    ut.ParamInfoBool('normonly_on', False, hideif=False),
                    ut.ParamInfoBool('bar_l2_on', False, hideif=False),
                    ut.ParamInfoBool('cos_on', False, hideif=False),
                    ut.ParamInfoBool('fg_on', True, hideif=False),
                    ut.ParamInfo(
                        'normalizer_rule', 'last', '', valid_values=['last', 'name']
                    ),
                    ut.ParamInfo(
                        'lnbnn_normer',
                        None,
                        hideif=None,
                        help_='config string for lnbnn score normalizer',
                    ),
                    ut.ParamInfo(
                        'lnbnn_norm_thresh',
                        0.5,
                        type_=float,
                        hideif=lambda cfg: not cfg['lnbnn_normer'],
                        help_='config string for lnbnn score normalizer',
                    ),
                    #
                    ut.ParamInfoBool('can_match_sameimg', False, 'sameimg', hideif=False),
                    ut.ParamInfoBool('can_match_samename', True, 'samename', hideif=True),
                    ut.ParamInfoBool('sqrd_dist_on', False, hideif=True),
                ],
            ]
        )
        return param_info_list


@six.add_metaclass(ConfigMetaclass)
class FeatureWeightConfig(ConfigBase):
    """

    CommandLine:
        python -m wbia.algo.Config --exec-FeatureWeightConfig

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.Config import *  # NOQA
        >>> featweight_cfg = FeatureWeightConfig(fw_detector='rf',
        >>>                                      featweight_enabled=True)
        >>> result = featweight_cfg.get_cfgstr()
        >>> print(result)

        _FEATWEIGHT(ON,uselabel,rf)_FEAT(hesaff+sift_)_CHIP(sz450)

        _FEATWEIGHT(OFF)_FEAT(hesaff+sift_)_CHIP(sz450)

    """

    @profile
    def __init__(featweight_cfg, **kwargs):
        super(FeatureWeightConfig, featweight_cfg).__init__(name='featweight_cfg')
        # Featweights depend on features
        featweight_cfg._feat_cfg = FeatureConfig(**kwargs)
        featweight_cfg.initialize_params()
        featweight_cfg.update(**kwargs)

    def get_param_info_list(self):
        from wbia import core_annots

        return (
            core_annots.ProbchipConfig._param_info_list
            + core_annots.FeatWeightConfig._param_info_list
        )


@six.add_metaclass(ConfigMetaclass)
class QueryConfig(ConfigBase):
    """
    LNBNN ranking query configuration parameters

    Example:
        >>> # ENABLE_DOCTEST
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> cfg = ibs.cfg.query_cfg
        >>> cfgstr = ibs.cfg.query_cfg.get_cfgstr()
        >>> print(cfgstr)
    """

    # TODO: make this a dtool Config
    _todo_subconfig_list = [
        NNConfig,
        NNWeightConfig,
        SpatialVerifyConfig,
        FlannConfig,
        FeatureWeightConfig,
    ]

    @profile
    def __init__(query_cfg, **kwargs):
        super(QueryConfig, query_cfg).__init__(name='query_cfg')
        query_cfg.nn_cfg = NNConfig(**kwargs)
        query_cfg.nnweight_cfg = NNWeightConfig(**kwargs)
        query_cfg.sv_cfg = SpatialVerifyConfig(**kwargs)
        query_cfg.agg_cfg = AggregateConfig(**kwargs)
        query_cfg.flann_cfg = FlannConfig(**kwargs)
        # causes some bug in Preference widget if these don't have underscore
        query_cfg._featweight_cfg = FeatureWeightConfig(**kwargs)
        query_cfg.use_cache = False
        # Start of pipeline
        query_cfg._valid_pipeline_roots = ['vsmany']
        query_cfg.pipeline_root = 'vsmany'
        # <Hack Paramaters>
        # query_cfg.with_metadata = False
        query_cfg.query_rotation_heuristic = False
        # query_cfg.query_rotation_heuristic = True
        query_cfg.codename = 'None'
        query_cfg.species_code = '____'  # TODO: make use of this
        # Depends on feature config
        query_cfg.update_query_cfg(**kwargs)
        if ut.VERYVERBOSE:
            logger.info('[config] NEW QueryConfig')

    def get_cfgstr_list(query_cfg, **kwargs):
        # Ensure feasibility of the configuration
        query_cfg.make_feasible()

        # Build cfgstr
        cfgstr_list = ['_' + query_cfg.pipeline_root]
        # if six.text_type(query_cfg.pipeline_root) == 'smk':
        #    # SMK Parameters
        #    if kwargs.get('use_smk', True):
        #        cfgstr_list += query_cfg.smk_cfg.get_cfgstr_list(**kwargs)
        #    if kwargs.get('use_sv', True):
        #        cfgstr_list += query_cfg.sv_cfg.get_cfgstr_list(**kwargs)
        if six.text_type(query_cfg.pipeline_root) == 'vsmany':
            # Naive Bayes Parameters
            if kwargs.get('use_nn', True):
                cfgstr_list += query_cfg.nn_cfg.get_cfgstr_list(**kwargs)
            if kwargs.get('use_nnweight', True):
                cfgstr_list += query_cfg.nnweight_cfg.get_cfgstr_list(**kwargs)
            if kwargs.get('use_sv', True):
                cfgstr_list += query_cfg.sv_cfg.get_cfgstr_list(**kwargs)
            if kwargs.get('use_agg', True):
                cfgstr_list += query_cfg.agg_cfg.get_cfgstr_list(**kwargs)
            if kwargs.get('use_flann', True):
                cfgstr_list += query_cfg.flann_cfg.get_cfgstr_list(**kwargs)
        else:
            raise AssertionError(
                'bad pipeline root: ' + six.text_type(query_cfg.pipeline_root)
            )
        if kwargs.get('use_featweight', True):
            cfgstr_list += query_cfg._featweight_cfg.get_cfgstr_list(**kwargs)
            # HACK: featweight_cfg used to include chip and feat
            # but they arent working now due to new structures, so they are hacked in here
            # This whole file will eventually be depricated
            cfgstr_list += query_cfg._featweight_cfg._feat_cfg.get_cfgstr_list(**kwargs)
            cfgstr_list += query_cfg._featweight_cfg._feat_cfg._chip_cfg.get_cfgstr_list(
                **kwargs
            )

        if query_cfg.query_rotation_heuristic:
            # HACK
            cfgstr_list += ['_HACK(augment_queryside)']
        return cfgstr_list

    def update_query_cfg(query_cfg, **cfgdict):
        # Each config paramater should be unique
        # So updating them all should not cause conflicts
        # FIXME: Should be able to infer all the children that need updates
        #
        # apply codename before updating subconfigs
        query_cfg.apply_codename(cfgdict.get('codename', None))
        # update subconfigs
        query_cfg.nn_cfg.update(**cfgdict)
        query_cfg.nnweight_cfg.update(**cfgdict)
        query_cfg.sv_cfg.update(**cfgdict)
        query_cfg.agg_cfg.update(**cfgdict)
        query_cfg.flann_cfg.update(**cfgdict)
        query_cfg._featweight_cfg.update(**cfgdict)
        query_cfg._featweight_cfg._feat_cfg.update(**cfgdict)
        query_cfg._featweight_cfg._feat_cfg._chip_cfg.update(**cfgdict)
        query_cfg.update(**cfgdict)
        # Ensure feasibility of the configuration
        try:
            query_cfg.make_feasible()
        except AssertionError as ex:
            logger.info(ut.repr2(cfgdict, sorted_=True))
            ut.printex(ex)
            raise

    def apply_codename(query_cfg, codename=None):
        """
        codenames denote mass changes to configurations
        it is a hacky solution to setting different parameter
        values all at once.
        """
        if codename is None:
            codename = query_cfg.codename

        # nnweight_cfg = query_cfg.nnweight_cfg
        # nn_cfg   = query_cfg.nn_cfg
        # agg_cfg = query_cfg.agg_cfg

        if codename.startswith('csum') or codename.endswith('_csum'):
            raise NotImplementedError('codename nsum')
        if codename.startswith('nsum'):
            raise NotImplementedError('codename nsum')
        if codename.startswith('vsmany'):
            query_cfg.pipeline_root = 'vsmany'
        elif codename.startswith('vsone'):
            assert False, 'no longer supporte'
        elif codename == 'None':
            pass

    def make_feasible(query_cfg):
        try:
            query_cfg.make_feasible_()
        except AssertionError as ex:
            if ut.NOT_QUIET:
                query_cfg.printme3()
            ut.printex(ex, 'failed ot make feasible')
            raise

    def make_feasible_(query_cfg):
        """
        removes invalid parameter settings over all cfgs (move to QueryConfig)
        """
        nnweight_cfg = query_cfg.nnweight_cfg
        nn_cfg = query_cfg.nn_cfg
        featweight_cfg = query_cfg._featweight_cfg
        # feat_cfg = query_cfg._featweight_cfg._feat_cfg
        # smk_cfg = query_cfg.smk_cfg
        # vocabassign_cfg = query_cfg.smk_cfg.vocabassign_cfg
        # agg_cfg = query_cfg.agg_cfg
        # sv_cfg = query_cfg.sv_cfg

        hasvalid_root = any(
            [
                query_cfg.pipeline_root.lower() == root.lower()
                for root in query_cfg._valid_pipeline_roots
            ]
        )
        try:
            assert hasvalid_root, 'invalid pipeline root %r valid roots are %r' % (
                query_cfg.pipeline_root,
                query_cfg._valid_pipeline_roots,
            )
        except AssertionError as ex:
            ut.printex(ex)
            raise

        # HACK
        if nnweight_cfg.fg_on is not True:
            featweight_cfg.featweight_enabled = False
        if featweight_cfg.featweight_enabled is not True:
            nnweight_cfg.fg_on = False

        nn_cfg.make_feasible()

    def deepcopy(query_cfg, **kwargs):
        copy_ = copy.deepcopy(query_cfg)
        copy_.update_query_cfg(**kwargs)
        return copy_


@six.add_metaclass(ConfigMetaclass)
class FeatureConfig(ConfigBase):
    """
    Feature configuration object.

    TODO depcirate for core_annots.FeatConfig

    CommandLine:
        python -m wbia.algo.Config --test-FeatureConfig

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo import Config  # NOQA
        >>> from wbia.algo.Config import *  # NOQA
        >>> feat_cfg = Config.FeatureConfig()
        >>> result = (feat_cfg.get_cfgstr())
        >>> print(result)
        >>> #assert result.startswith('_FEAT(hesaff+sift_)_CHIP')
        _Feat(hesaff+sift)
    """

    def __init__(feat_cfg, **kwargs):
        # Features depend on chips
        # import pyhesaff
        super(FeatureConfig, feat_cfg).__init__(name='feat_cfg')
        feat_cfg._chip_cfg = ChipConfig(**kwargs)
        feat_cfg.initialize_params()
        # feat_cfg.feat_type = 'hesaff+sift'
        # feat_cfg.bgmethod = None
        # feat_cfg._param_list = list(six.iteritems(
        #    pyhesaff.get_hesaff_default_params()))
        # for type_, name, default, doc in feat_cfg._iterparams():
        #    setattr(feat_cfg, name, default)
        # feat_cfg.use_adaptive_scale = False  # 9001 # 80
        # feat_cfg.nogravity_hack = False  # 9001 # 80
        feat_cfg.update(**kwargs)

    def get_param_info_list(self):
        from wbia import core_annots

        return core_annots.FeatConfig().get_param_info_list()

    def get_config_name(self):
        return 'Feat'

    def get_hesaff_params(feat_cfg):
        import pyhesaff

        default_keys = list(pyhesaff.get_hesaff_default_params().keys())
        hesaff_param_dict = ut.dict_subset(feat_cfg, default_keys)
        return hesaff_param_dict


@six.add_metaclass(ConfigMetaclass)
class ChipConfig(ConfigBase):
    """ ChipConfig """

    def __init__(cc_cfg, **kwargs):
        super(ChipConfig, cc_cfg).__init__(name='chip_cfg')
        cc_cfg.initialize_params()
        # cc_cfg.dim_size    = 450
        # # cc_cfg.resize_dim  = 'area'
        # cc_cfg.resize_dim  = 'width'
        # cc_cfg.grabcut     = False
        # cc_cfg.histeq      = False
        # cc_cfg.adapteq     = False
        # cc_cfg.region_norm = False
        # cc_cfg.rank_eq     = False
        # cc_cfg.local_eq    = False
        # cc_cfg.maxcontrast = False
        # cc_cfg.chipfmt     = '.png'
        cc_cfg.update(**kwargs)

    def get_param_info_list(self):
        from wbia import core_annots

        return core_annots.ChipConfig._param_info_list


@six.add_metaclass(ConfigMetaclass)
class DetectionConfig(ConfigBase):
    """
    CommandLine:
        python -m wbia.algo.Config --test-DetectionConfig

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.Config import *  # NOQA
        >>> detect_cfg = DetectionConfig()
        >>> result = (detect_cfg.get_cfgstr())
        >>> print(result)
        _DETECT(cnn,____,sz=800)
    """

    def __init__(detect_cfg, **kwargs):
        super(DetectionConfig, detect_cfg).__init__(name='detect_cfg')
        # detect_cfg.species_text = 'zebra_grevys'
        detect_cfg.species_text = const.UNKNOWN
        # detect_cfg.detector = 'rf'
        detect_cfg.detector = 'cnn'
        detect_cfg.scale_list = '1.25, 1.0, 0.80, 0.65, 0.50, 0.40, 0.30, 0.20, 0.10'
        detect_cfg.trees_path = ''
        detect_cfg.detectimg_sqrt_area = 800
        detect_cfg.update(**kwargs)

    def get_cfgstr_list(detect_cfg):
        cfgstrs = [
            '_DETECT(',
            detect_cfg.detector,
            ',',
            detect_cfg.species_text,
            ',sz=%d' % (detect_cfg.detectimg_sqrt_area,),
            ')',
        ]
        return cfgstrs


@six.add_metaclass(ConfigMetaclass)
class OccurrenceConfig(ConfigBase):
    """OccurrenceConfig

    CommandLine:
        python -m wbia.algo.Config --exec-OccurrenceConfig --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.Config import *  # NOQA
        >>> occur_cfg = OccurrenceConfig()
        >>> print(occur_cfg.get_cfgstr())
    """

    def __init__(occur_cfg, **kwargs):
        super(OccurrenceConfig, occur_cfg).__init__(name='occur_cfg')
        occur_cfg.initialize_params()
        occur_cfg.update(**kwargs)

    def get_param_info_list(occur_cfg):
        param_info_list = [
            ut.ParamInfo('min_imgs_per_occurrence', 1, 'minper='),
            # ut.ParamInfo('cluster_algo', 'agglomerative', '', valid_values=['agglomerative', 'meanshift']),
            ut.ParamInfo(
                'cluster_algo', 'agglomerative', '', valid_values=['agglomerative']
            ),
            # ut.ParamInfo('quantile', .01, 'quant', hideif=lambda cfg: cfg['cluster_algo'] != 'meanshift'),
            ut.ParamInfo(
                'seconds_thresh',
                1600,
                'sec',
                hideif=lambda cfg: cfg['cluster_algo'] != 'agglomerative',
            ),
            ut.ParamInfo('use_gps', True, hideif=False),
            ut.ParamInfo('km_per_sec', 0.002),
        ]
        return param_info_list


@six.add_metaclass(ConfigMetaclass)
class DisplayConfig(ConfigBase):
    """ DisplayConfig """

    def __init__(display_cfg, **kwargs):
        super(DisplayConfig, display_cfg).__init__(name='display_cfg')
        display_cfg.N = 6
        display_cfg.name_scoring = False
        display_cfg.showanalysis = False
        display_cfg.annotations = True
        display_cfg.vert = True  # None
        display_cfg.show_results_in_image = False  # None

    def get_cfgstr_list(nn_cfg):
        raise NotImplementedError('abstract')
        return ['unimplemented']


@six.add_metaclass(ConfigMetaclass)
class OtherConfig(ConfigBase):
    def __init__(other_cfg, **kwargs):
        super(OtherConfig, other_cfg).__init__(name='other_cfg')
        # other_cfg.thumb_size      = 128
        other_cfg.thumb_size = 221
        other_cfg.thumb_bare_size = 700
        other_cfg.ranks_top = 2
        other_cfg.filter_reviewed = True
        other_cfg.auto_localize = True
        # maximum number of exemplars per name
        other_cfg.max_exemplars = 6
        other_cfg.exemplars_per_view = 2
        other_cfg.prioritized_subset_annots_per_name = 2
        other_cfg.exemplar_distinctiveness_thresh = 0.95
        other_cfg.detect_add_after = 1
        # other_cfg.detect_use_chunks = True
        other_cfg.use_more_special_imagesets = True
        other_cfg.location_for_names = 'IBEIS'
        # other_cfg.location_for_names = 'MUGU'
        other_cfg.smart_enabled = True
        other_cfg.enable_custom_filter = False
        other_cfg.hots_batch_size = 256
        other_cfg.use_augmented_indexer = True
        other_cfg.show_shipped_imagesets = ut.is_developer()
        other_cfg.update(**kwargs)


# Convinience
def __dict_default_func(dict_):
    # Sets keys only if they dont exist
    def set_key(key, val):
        if key not in dict_:
            dict_[key] = val

    return set_key


def default_vsone_cfg(ibs, **kwargs):
    # DEPRICATE
    kwargs['pipeline_root'] = 'vsone'
    ut.dict_update_newkeys(
        kwargs,
        {
            'lnbnn_on': False,
            'checks': 256,
            'K': 1,
            'Knorm': 1,
            'ratio_thresh': 0.6666,  # 1.5,
        },
    )
    query_cfg = QueryConfig(**kwargs)
    return query_cfg


def set_query_cfg(cfg, query_cfg):
    """ hack 12-30-2014 """
    cfg.query_cfg = query_cfg
    cfg.featweight_cfg = cfg.query_cfg._featweight_cfg
    cfg.feat_cfg = cfg.query_cfg._featweight_cfg._feat_cfg
    cfg.chip_cfg = cfg.query_cfg._featweight_cfg._feat_cfg._chip_cfg


def update_query_config(cfg, **kwargs):
    """ hack 12-30-2014 """
    cfg.query_cfg.update_query_cfg(**kwargs)
    cfg.featweight_cfg = cfg.query_cfg._featweight_cfg
    cfg.feat_cfg = cfg.query_cfg._featweight_cfg._feat_cfg
    cfg.chip_cfg = cfg.query_cfg._featweight_cfg._feat_cfg._chip_cfg


@profile
def load_named_config(
    cfgname, dpath, use_config_cache=False, verbose=ut.VERBOSE and ut.NOT_QUIET
):
    """hack 12-30-2014

    Args:
        cfgname (str):
        dpath (str):
        use_config_cache (bool):

    Returns:
        Config: cfg

    CommandLine:
        python -m wbia.algo.Config --test-load_named_config

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.Config import *  # NOQA
        >>> from wbia.algo.Config import _default_config  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('PZ_Master0')
        >>> #ibs.cfg.save()
        >>> # build test data
        >>> cfgname = 'zebra_plains'
        >>> dpath = ibs.get_dbdir()
        >>> use_config_cache = True
        >>> # execute function
        >>> cfg = load_named_config(cfgname, dpath, use_config_cache)
        >>> #
        >>> keys1 = ut.take_column(cfg.parse_items(), 0)
        >>> keys2 = ut.take_column(ibs.cfg.parse_items(), 0)
        >>> symdiff = set(keys1) ^ set(keys2)
        >>> # verify results
        >>> result = six.text_type(cfg)
        >>> print(result)
    """
    if cfgname is None:
        # TODO: find last cfgname
        cfgname = 'cfg'
    fpath = join(dpath, cfgname) + '.cPkl'
    if verbose:
        logger.info('[Config] loading named config fpath=%r' % (fpath,))
    # Always a fresh object
    cfg = GenericConfig(cfgname, fpath=fpath)
    try:
        # Use pref cache
        if not use_config_cache:
            raise Exception('force config cache miss')
        # Get current "schema"
        tmp = _default_config(cfg, cfgname, new=True)
        current_itemset = tmp.parse_items()
        current_keyset = list(ut.take_column(current_itemset, 0))
        # load saved preferences
        cfg.load()
        # Check if loaded schema has changed
        loaded_keyset = list(ut.take_column(cfg.parse_items(), 0))
        missing_keys = set(current_keyset) - set(loaded_keyset)
        if len(missing_keys) != 0:
            # Bring over new values into old structure
            tmp.update(**dict(cfg.parse_items()))
            cfg = tmp
            # missing_vals = ut.dict_take(dict(current_itemset), missing_keys)
            # def find_cfgkey_parent(tmp, key):
            #    subconfig_list = []
            #    for attr in dir(tmp):
            #        if attr == key:
            #            return tmp
            #        child = getattr(tmp, attr)
            #        if isinstance(child, ConfigBase):
            #            subconfig_list.append(child)
            #    for subconfig in subconfig_list:
            #        found = find_cfgkey_parent(subconfig, key)
            #        if found is not None:
            #            return found
            # missing_parents = [find_cfgkey_parent(tmp, key) for key in missing_keys]
            # for parent, key, val in zip(missing_parents, missing_keys, missing_vals):
            #    setattr(parent, key, val)
        #    # TODO: Finishme update the out of data preferences
        #    pass
        if verbose:
            logger.info('[Config] successfully loaded config cfgname=%r' % (cfgname,))
    except Exception as ex:
        if ut.VERBOSE:
            ut.printex(ex, iswarning=True)
        # Totally new completely default preferences
        cfg = _default_config(cfg, cfgname)
        cfg.save()
    # Hack in cfgname
    if verbose:
        logger.info('[Config] hack in z_cfgname=%r' % (cfgname,))
    cfg.z_cfgname = cfgname
    return cfg


@profile
def _default_config(cfg, cfgname=None, new=True):
    """ hack 12-30-2014 """
    if ut.VERBOSE:
        logger.info('[Config] building default config')
    if cfgname is None:
        cfgname = cfg.z_cfgname
    if new:
        fpath = cfg.get_fpath()
        cfg = GenericConfig(cfgname, fpath=fpath)
        cfg.z_cfgname = cfgname
    query_cfg = QueryConfig(pipeline_root='vsmany')
    set_query_cfg(cfg, query_cfg)
    cfg.occur_cfg = OccurrenceConfig()
    cfg.detect_cfg = DetectionConfig()
    cfg.other_cfg = OtherConfig()
    _default_named_config(cfg, cfgname)
    # if len(species_list) == 1:
    #    # try to be intelligent about the default speceis
    #    cfg.detect_cfg.species_text = species_list[0]
    return cfg


@profile
def _default_named_config(cfg, cfgname):
    """hack 12-30-2014

    list default parameters per species

    """
    if cfgname == 'cfg':
        cfg.detect_cfg.species_text = 'none'
    elif cfgname == 'zebra_plains':
        cfg.detect_cfg.species_text = cfgname
        # speedup': 46.90769958496094,
        cfg.query_cfg.flann_cfg.algorithm = 'kdtree'
        cfg.query_cfg.flann_cfg.trees = 8
        cfg.query_cfg.nn_cfg.checks = 704
        # 'algorithm': 'kdtree',
        # [dev.tune_flann]    'checks': 6656,
        # [dev.tune_flann]    'trees': 4,

        # Kmeans seems a bit more accurate
        # 'algorithm': 'kmeans',
        # 'branching': 16,
        # 'cb_index': 0.6000000238418579,
        # 'centers_init': 'random',
        # 'checks': 18432,
        # 'iterations': 1,
        # 'leaf_max_size': 4,
        # 'speedup': 65.54280090332031,
        # 'target_precision': 0.9800000190734863,

    elif cfgname == 'zebra_grevys':
        cfg.detect_cfg.species_text = cfgname
        # speedup': 224.7425994873047,
        cfg.query_cfg.flann_cfg.algorithm = 'kdtree'
        cfg.query_cfg.flann_cfg.trees = 4
        cfg.query_cfg.nn_cfg.checks = 896
    elif cfgname == 'giraffe_reticulated':
        cfg.detect_cfg.species_text = cfgname
        cfg.query_cfg.flann_cfg.algorithm = 'kdtree'
        cfg.query_cfg.flann_cfg.trees = 8
        cfg.query_cfg.nn_cfg.checks = 316
    else:
        if ut.VERBOSE:
            logger.info('WARNING: UNKNOWN CFGNAME=%r' % (cfgname,))



# -*- coding: utf-8 -*-
import logging
import six

# import six
import utool as ut
from wbia import guitool

print, rrr, profile = ut.inject2(__name__, '[expt_harn]')
logger = logging.getLogger('wbia')


@six.add_metaclass(ut.ReloadingMetaclass)
class ResultMetadata(object):
    def __init__(metadata, fpath, autoconnect=False):
        """
        metadata_fpath = join(figdir, 'result_metadata.shelf')
        metadata = ResultMetadata(metadata_fpath)
        """
        metadata.fpath = fpath
        metadata.dimensions = ['query_cfgstr', 'qaid']
        metadata._shelf = None
        metadata.dictstore = None
        if autoconnect:
            metadata.connect()

    def connect(metadata):
        import shelve

        metadata._shelf = shelve.open(metadata.fpath)
        if 'dictstore' not in metadata._shelf:
            dictstore = ut.AutoVivification()
            metadata._shelf['dictstore'] = dictstore
        metadata.dictstore = metadata._shelf['dictstore']

    def clear(metadata):
        dictstore = ut.AutoVivification()
        metadata._shelf['dictstore'] = dictstore
        metadata.dictstore = metadata._shelf['dictstore']

    def __del__(metadata):
        metadata.close()

    def close(metadata):
        metadata._shelf.close()

    def write(metadata):
        metadata._shelf['dictstore'] = metadata.dictstore
        metadata._shelf.sync()

    def set_global_data(metadata, cfgstr, qaid, key, val):
        metadata.dictstore[cfgstr][qaid][key] = val

    def sync_test_results(metadata, testres):
        """ store all test results in the shelf """
        for cfgx in range(len(testres.cfgx2_qreq_)):
            cfgstr = testres.get_cfgstr(cfgx)
            qaids = testres.qaids
            cfgresinfo = testres.cfgx2_cmsinfo[cfgx]
            for key, val_list in six.iteritems(cfgresinfo):
                for qaid, val in zip(qaids, val_list):
                    metadata.set_global_data(cfgstr, qaid, key, val)
        metadata.write()

    def get_cfgstr_list(metadata):
        cfgstr_list = list(metadata.dictstore.keys())
        return cfgstr_list

    def get_column_keys(metadata):
        unflat_colname_list = [
            [cols.keys() for cols in qaid2_cols.values()]
            for qaid2_cols in six.itervalues(metadata.dictstore)
        ]
        colname_list = ut.unique_ordered(ut.flatten(ut.flatten(unflat_colname_list)))
        return colname_list

    def get_square_data(metadata, cfgstr=None):
        # can only support one config at a time right now
        if cfgstr is None:
            cfgstr = metadata.get_cfgstr_list()[0]
        qaid2_cols = metadata.dictstore[cfgstr]
        qaids = list(qaid2_cols.keys())
        col_name_list = ut.unique_ordered(
            ut.flatten([cols.keys() for cols in qaid2_cols.values()])
        )
        # col_name_list = ['qx2_scoreexpdiff', 'qx2_gt_aid']
        # colname2_colvals = [None for colname in col_name_list]
        column_list = [
            [colvals.get(colname, None) for qaid, colvals in six.iteritems(qaid2_cols)]
            for colname in col_name_list
        ]
        col_name_list = ['qaids'] + col_name_list
        column_list = [qaids] + column_list
        logger.info('depth_profile(column_list) = %r' % (ut.depth_profile(column_list),))
        return col_name_list, column_list


def make_metadata_custom_api(metadata):
    r"""
    CommandLine:
        python -m wbia.expt.experiment_drawing --test-make_metadata_custom_api --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.expt.experiment_drawing import *  # NOQA
        >>> import wbia.guitool
        >>> guitool.ensure_qapp()
        >>> metadata_fpath = '/media/raid/work/Elephants_drop1_ears/_ibsdb/figures/result_metadata.shelf'
        >>> metadata = ResultMetadata(metadata_fpath, autoconnect=True)
        >>> wgt = make_metadata_custom_api(metadata)
        >>> ut.quit_if_noshow()
        >>> wgt.show()
        >>> wgt.raise_()
        >>> guitool.qtapp_loop(wgt, frequency=100)
    """
    from wbia.guitool.__PYQT__ import QtCore

    class MetadataViewer(guitool.APIItemWidget):
        def __init__(wgt, parent=None, tblnice='Result Metadata Viewer', **kwargs):
            guitool.APIItemWidget.__init__(wgt, parent=parent, tblnice=tblnice, **kwargs)
            wgt.connect_signals_and_slots()

        @guitool.slot_(QtCore.QModelIndex)
        def _on_doubleclick(wgt, qtindex):
            logger.info('[wgt] _on_doubleclick: ')
            col = qtindex.column()
            if wgt.api.col_edit_list[col]:
                logger.info('do nothing special for editable columns')
                return
            model = qtindex.model()
            colname = model.get_header_name(col)
            if colname.endswith('fpath'):
                logger.info('showing fpath')
                fpath = model.get_header_data(colname, qtindex)
                ut.startfile(fpath)

        def connect_signals_and_slots(wgt):
            # wgt.view.clicked.connect(wgt._on_click)
            wgt.view.doubleClicked.connect(wgt._on_doubleclick)
            # wgt.view.pressed.connect(wgt._on_pressed)
            # wgt.view.activated.connect(wgt._on_activated)

    guitool.ensure_qapp()
    # cfgstr_list = metadata
    col_name_list, column_list = metadata.get_square_data()

    # Priority of column names
    colname_priority = [
        'qaids',
        'qx2_gt_rank',
        'qx2_gt_timedelta',
        'qx2_gf_timedelta',
        'analysis_fpath',
        'qx2_gt_raw_score',
        'qx2_gf_raw_score',
    ]
    colname_priority += sorted(ut.setdiff_ordered(col_name_list, colname_priority))
    sortx = ut.priority_argsort(col_name_list, colname_priority)
    col_name_list = ut.take(col_name_list, sortx)
    column_list = ut.take(column_list, sortx)

    col_lens = list(map(len, column_list))
    logger.info('col_name_list = %r' % (col_name_list,))
    logger.info('col_lens = %r' % (col_lens,))
    assert len(col_lens) > 0, 'no columns'
    assert col_lens[0] > 0, 'no rows'
    assert all([len_ == col_lens[0] for len_ in col_lens]), 'inconsistant data'
    col_types_dict = {}
    col_getter_dict = dict(zip(col_name_list, column_list))
    col_bgrole_dict = {}
    col_ider_dict = {}
    col_setter_dict = {}
    col_nice_dict = {name: name.replace('qx2_', '') for name in col_name_list}
    col_nice_dict.update(
        {
            'qx2_gt_timedelta': 'GT TimeDelta',
            'qx2_gf_timedelta': 'GF TimeDelta',
            'qx2_gt_rank': 'GT Rank',
        }
    )
    editable_colnames = []
    sortby = 'qaids'

    def get_thumb_size():
        return 128

    col_width_dict = {}
    custom_api = guitool.CustomAPI(
        col_name_list,
        col_types_dict,
        col_getter_dict,
        col_bgrole_dict,
        col_ider_dict,
        col_setter_dict,
        editable_colnames,
        sortby,
        get_thumb_size,
        sort_reverse=True,
        col_width_dict=col_width_dict,
        col_nice_dict=col_nice_dict,
    )
    # headers = custom_api.make_headers(tblnice='results')
    # logger.info(ut.repr2(headers))
    wgt = MetadataViewer()
    wgt.connect_api(custom_api)
    return wgt


def make_test_result_custom_api(ibs, testres):

    guitool.ensure_qapp()
    cfgx = 0
    cfgres_info = testres.cfgx2_cmsinfo[cfgx]
    qaids = testres.qaids
    gt_aids = cfgres_info['qx2_gt_aid']
    gf_aids = cfgres_info['qx2_gf_aid']
    qx2_gt_timedelta = ibs.get_annot_pair_timedelta(qaids, gt_aids)
    qx2_gf_timedelta = ibs.get_annot_pair_timedelta(qaids, gf_aids)
    col_name_list = [
        'qaids',
        'qx2_gt_aid',
        'qx2_gf_aid',
        'qx2_gt_timedelta',
        'qx2_gf_timedelta',
    ]
    col_types_dict = {}
    col_getter_dict = {}
    col_getter_dict.update(**cfgres_info)
    col_getter_dict['qaids'] = testres.qaids
    col_getter_dict['qx2_gt_timedelta'] = qx2_gt_timedelta
    col_getter_dict['qx2_gf_timedelta'] = qx2_gf_timedelta
    col_bgrole_dict = {}
    col_ider_dict = {}
    col_setter_dict = {}
    editable_colnames = []
    sortby = 'qaids'

    def get_thumb_size():
        return 128

    col_width_dict = {}

    custom_api = guitool.CustomAPI(
        col_name_list,
        col_types_dict,
        col_getter_dict,
        col_bgrole_dict,
        col_ider_dict,
        col_setter_dict,
        editable_colnames,
        sortby,
        get_thumb_size,
        True,
        col_width_dict,
    )
    # headers = custom_api.make_headers(tblnice='results')
    # logger.info(ut.repr2(headers))
    wgt = guitool.APIItemWidget()
    wgt.connect_api(custom_api)
    return wgt


@profile
def draw_results(ibs, testres):
    r"""
    Draws results from an experiment harness run.
    Rows store different qaids (query annotation ids)
    Cols store different configurations (algorithm parameters)

    Args:
        testres (TestResult):

    CommandLine:
        python dev.py -t custom:rrvsone_on=True,constrained_coeff=0 custom --qaid 12 --db PZ_MTEST --show --va
        python dev.py -t custom:rrvsone_on=True,constrained_coeff=.3 custom --qaid 12 --db PZ_MTEST --show --va --noqcache
        python dev.py -t custom:rrvsone_on=True custom --qaid 4 --db PZ_MTEST --show --va --noqcache
        python dev.py -t custom:rrvsone_on=True,grid_scale_factor=1 custom --qaid 12 --db PZ_MTEST --show --va --noqcache
        python dev.py -t custom:rrvsone_on=True,grid_scale_factor=1,grid_steps=1 custom --qaid 12 --db PZ_MTEST --show --va --noqcache

    CommandLine:
        python dev.py -t best --db seals2 --allgt --vz --fig-dname query_analysis_easy --show
        python dev.py -t best --db seals2 --allgt --vh --fig-dname query_analysis_hard --show

        python dev.py -t pyrscale --db PZ_MTEST --allgt --vn --fig-dname query_analysis_interesting --show
        python dev.py -t pyrscale --db testdb3 --allgt --vn --fig-dname query_analysis_interesting --vf
        python dev.py -t pyrscale --db testdb3 --allgt --vn --fig-dname query_analysis_interesting --vf --quality

        python -m wbia.expt.experiment_drawing --test-draw_results --show --vn
        python -m wbia.expt.experiment_drawing --test-draw_results --show --vn --db PZ_MTEST
        python -m wbia.expt.old_storage --test-draw_results --show --db PZ_MTEST --gv

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.expt.old_storage import *  # NOQA
        >>> from wbia.init import main_helpers
        >>> ibs, testres = main_helpers.testdata_expts('PZ_MTEST')
        >>> result = draw_results(ibs, testres)
        >>> # verify results
        >>> print(result)
    """
    logger.info(' --- DRAW RESULTS ---')

    # It is very inefficient to turn off caching when view_all is true

    figdir = ibs.get_fig_dir()
    ut.ensuredir(figdir)

    if ut.get_argflag(('--view-fig-directory', '--vf')):
        ut.view_directory(figdir)

    figdir_suffix = ut.get_argval('--fig-dname', type_=str, default=None)
    from os.path import join

    if figdir_suffix is not None:
        figdir = join(figdir, figdir_suffix)
        ut.ensuredir(figdir)
    # gx2_gt_timedelta
    #    cfgres_info['qx2_gf_timedelta'] = qx2_gf_timedelta

    metadata_fpath = join(figdir, 'result_metadata.shelf')
    metadata = ResultMetadata(metadata_fpath)
    # metadata.rrr()
    metadata.connect()
    metadata.sync_test_results(testres)
    # cfgstr = qreq_.get_cfgstr()
    # cfg_metadata = ensure_item(metadata, cfgstr, {})
    # avuuids = ibs.get_annot_visual_uuids(qaids)
    # avuuid2_ax = ensure_item(cfg_metadata, 'avuuid2_ax', {})
    # cfg_columns = ensure_item(cfg_metadata, 'columns', {})
    # import wbia.guitool

    # ut.argv_flag_dec(draw_rank_cmc)(ibs, testres)

    # VIZ_INDIVIDUAL_RESULTS = True
    # if VIZ_INDIVIDUAL_RESULTS:
    #     draw_match_cases(ibs, testres, metadata=metadata)

    metadata.write()
    if ut.get_argflag(('--guiview', '--gv')):
        guitool.ensure_qapp()
        # wgt = make_test_result_custom_api(ibs, testres)
        wgt = make_metadata_custom_api(metadata)
        wgt.show()
        wgt.raise_()
        guitool.qtapp_loop(wgt, frequency=100)
    metadata.close()

    if ut.NOT_QUIET:
        logger.info('[DRAW_RESULT] EXIT EXPERIMENT HARNESS')



# -*- coding: utf-8 -*-
# flake8:noqa
import utool as ut

ut.noinject(__name__, '[__PYQT__.__init__]')
from . import _internal

GUITOOL_PYQT_VERSION = _internal.GUITOOL_PYQT_VERSION

if _internal.GUITOOL_PYQT_VERSION == 4:
    from PyQt4 import *
elif _internal.GUITOOL_PYQT_VERSION == 5:
    from PyQt5 import *
else:
    raise ValueError('Unknown version of PyQt')

from . import QtCore
from . import QtGui
from . import QtWidgets  # just a clone of QtGui in PyQt4


def QVariantHack(*args):
    """ Hack when sip.setapi('QVariant') is 2 """
    if len(args) == 0:
        return None
    elif len(args) == 1:
        return args[0]
    else:
        raise NotImplementedError(str(args))


try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:

    def _fromUtf8(s):
        return s


try:
    _encoding = QtWidgets.QApplication.UnicodeUTF8

    def _translate(context, text, disambig):
        return QtWidgets.QApplication.translate(context, text, disambig, _encoding)


except AttributeError:
    _encoding = ut.identity

    def _translate(context, text, disambig):
        return QtWidgets.QApplication.translate(context, text, disambig)


# print('__pyqt5__2')



# -*- coding: utf-8 -*-
import logging
import torch
from collections import defaultdict
from os.path import join
from torch.autograd import Variable
from wbia.algo.verif.torch import netmath
import tensorboard_logger
import utool as ut

# from wbia.algo.verif.torch import gpu_util

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


class FitHarness(object):
    def __init__(
        harn,
        model,
        train_loader,
        vali_loader=None,
        test_loader=None,
        criterion='cross_entropy',
        lr_scheduler='exp',
        optimizer_cls='Adam',
        class_weights=None,
        gpu_num=None,
        workdir=None,
    ):

        harn.workdir = workdir

        harn.train_loader = train_loader
        harn.vali_loader = vali_loader
        harn.test_loader = test_loader

        harn.model = model

        harn.optimizer_cls = optimizer_cls
        harn.criterion = criterion
        harn.lr_scheduler = lr_scheduler
        # netmath.Optimizers.lookup(optimizer_cls)
        # netmath.Criterions.lookup(criterion)
        # netmath.LRSchedules.lookup(lr_scheduler)

        harn.class_weights = class_weights

        harn.gpu_num = gpu_num
        harn.use_cuda = harn.gpu_num is not None

        # harn.model = torch.nn.DataParallel(model, device_ids=[0, 1]).cuda()
        harn.config = {
            'maxIterations': 10000,
            'displayInterval': 1,
            'vail_displayInterval': 1,
            'model_dir': '.',
            'margin': 1.0,
        }
        harn.lr = harn.lr_scheduler.init_lr
        harn.epoch = 0

    def log(harn, msg):
        logger.info(msg)

    def log_value(harn, key, value, n_iter):
        if False:
            logger.info('{}={} @ {}'.format(key, value, n_iter))
        if tensorboard_logger:
            tensorboard_logger.log_value(key, value, n_iter)

    def _to_xpu(harn, *args):
        """ Puts data on the GPU if available """
        if harn.use_cuda:
            args = [Variable(item.cuda(harn.gpu_num)) for item in args]
            # input_batch = [Variable(item.cuda()) for item in input_batch]
        else:
            args = [Variable(item) for item in args]
        return args

    def run(harn):
        harn.log('Begin training')

        if False:
            # TODO: can we run this as a subprocess that dies when we die?
            # or do we need to run externally?
            # tensorboard --logdir runs
            # http://aretha:6006
            pass

        if tensorboard_logger:
            harn.log('Initializing tensorboard')
            tensorboard_logger.configure('runs/wbia', flush_secs=2)

        if harn.use_cuda:
            harn.log('Fitting model on GPU({})'.format(harn.gpu_num))
            harn.model.cuda(harn.gpu_num)
        else:
            harn.log('Fitting model on the CPU')

        if harn.class_weights is not None:
            (harn.class_weights,) = harn._to_xpu(harn.class_weights)

        lr = harn.lr_scheduler(harn.epoch)
        harn.optimizer = harn.optimizer_cls(harn.model.parameters(), lr=lr)

        # train loop

        while not harn.check_termination():
            harn.train_epoch()

            if harn.vali_loader:
                harn.validation_epoch()

            harn.save_snapshot()

            harn.epoch += 1

    def check_termination(harn):
        # check for termination
        if harn.epoch > harn.config['maxIterations']:
            harn.log('Maximum harn.epoch reached, terminating ...')
            return True
        return False

    def train_epoch(harn):
        ave_metrics = defaultdict(lambda: 0)

        # change learning rate (modified optimizer inplace)
        lr = harn.lr_scheduler(harn.epoch, harn.optimizer)

        # train batch
        for batch_idx, input_batch in enumerate(harn.train_loader):
            input_batch = harn._to_xpu(*input_batch)

            # logger.info('Begin batch {}'.format(batch_idx))
            t_cur_metrics = harn.train_batch(input_batch)

            for k, v in t_cur_metrics.items():
                ave_metrics[k] += v

            # display training info
            if (batch_idx + 1) % harn.config['displayInterval'] == 0:
                for k in ave_metrics.keys():
                    ave_metrics[k] /= harn.config['displayInterval']

                n_train = len(harn.train_loader)
                harn.log(
                    'Epoch {0}: {1} / {2} | lr:{3} - tloss:{4:.5f} acc:{5:.2f} | sdis:{6:.3f} ddis:{7:.3f}'.format(
                        harn.epoch,
                        batch_idx,
                        n_train,
                        lr,
                        ave_metrics['loss'],
                        ave_metrics['accuracy'],
                        ave_metrics['pos_dist'],
                        ave_metrics['neg_dist'],
                    )
                )

                iter_idx = harn.epoch * n_train + batch_idx
                for key, value in ave_metrics.items():
                    harn.log_value('train ' + key, value, iter_idx)

                # diagnoseGradients(model.parameters())
                for k in ave_metrics.keys():
                    ave_metrics[k] = 0

    def validation_epoch(harn):
        ave_metrics = defaultdict(lambda: 0)

        final_metrics = ave_metrics.copy()

        for vali_idx, input_batch in enumerate(harn.vali_loader):
            input_batch = harn._to_xpu(*input_batch)

            # logger.info('Begin batch {}'.format(vali_idx))
            v_cur_metrics = harn.validation_batch(input_batch)

            for k, v in v_cur_metrics.items():
                ave_metrics[k] += v
                final_metrics[k] += v

            if (vali_idx + 1) % harn.config['vail_displayInterval'] == 0:
                for k in ave_metrics.keys():
                    ave_metrics[k] /= harn.config['displayInterval']

                harn.log(
                    'Epoch {0}: {1} / {2} | vloss:{3:.5f} acc:{4:.2f} | sdis:{5:.3f} ddis:{6:.3f}'.format(
                        harn.epoch,
                        vali_idx,
                        len(harn.vali_loader),
                        ave_metrics['loss'],
                        ave_metrics['accuracy'],
                        ave_metrics['pos_dist'],
                        ave_metrics['neg_dist'],
                    )
                )

                for k in ave_metrics.keys():
                    ave_metrics[k] = 0

        for k in final_metrics.keys():
            final_metrics[k] /= len(harn.vali_loader)
        harn.log(
            'Epoch {0}: final vloss:{1:.5f} acc:{2:.2f} | sdis:{3:.3f} ddis:{4:.3f}'.format(
                harn.epoch,
                final_metrics['loss'],
                final_metrics['accuracy'],
                final_metrics['pos_dist'],
                final_metrics['neg_dist'],
            )
        )

        iter_idx = harn.epoch * len(harn.vali_loader) + vali_idx
        for key, value in final_metrics.items():
            harn.log_value('validation ' + key, value, iter_idx)

    # def display_metrics():
    #     pass

    def load_snapshot(harn, load_path):
        snapshot = torch.load(load_path)
        harn.model.load_state_dict(snapshot['model_state_dict'])
        harn.epoch = snapshot['epoch']
        harn.log('Model loaded from {}'.format(load_path))

    def save_snapshot(harn):
        # save snapshot
        save_path = join(
            harn.config['model_dir'], 'snapshot_epoch_{}.pt'.format(harn.epoch)
        )
        snapshot = {
            'epoch': harn.epoch,
            'model_state_dict': harn.model.state_dict(),
        }
        torch.save(snapshot, save_path)
        harn.log('Snapshot saved to {}'.format(save_path))

    def train_batch(harn, input_batch):
        """
        https://github.com/meetshah1995/pytorch-semseg/blob/master/train.py
        """
        harn.model.train(True)
        *inputs, label = input_batch

        # Forward prop through the model
        output = harn.model(*inputs)

        # Compute the loss
        loss = harn.criterion(output, label, weight=harn.class_weights)

        # Measure train accuracy and other informative metrics
        t_metrics = harn._measure_metrics(output, label, loss)

        # Backprop and learn
        harn.optimizer.zero_grad()
        loss.backward()
        harn.optimizer.step()

        return t_metrics

    def validation_batch(harn, input_batch):
        harn.model.train(False)
        *inputs, label = input_batch

        output = harn.model(*inputs)

        loss = harn.criterion(output, label, weight=harn.class_weights)

        # Measure validation accuracy and other informative metrics
        v_metrics = harn._measure_metrics(output, label, loss)

        return v_metrics

    def _measure_metrics(harn, output, label, loss):
        metrics = netmath.Metrics._siamese_metrics(
            output, label, margin=harn.criterion.margin
        )

        assert 'loss' not in metrics, 'cannot compute loss as an extra metric'

        loss_sum = loss.data.sum()
        inf = float('inf')
        if loss_sum == inf or loss_sum == -inf:
            harn.log('WARNING: received an inf loss, setting loss value to 0')
            loss_value = 0
        else:
            loss_value = loss.data[0]

        metrics['loss'] = loss_value
        # metrics = {
        #     'tpr': netmath.Metrics.tpr(output, label)
        # }
        return metrics



# -*- coding: utf-8 -*-
"""Developer convenience functions for ibs (detections).

TODO: need to split up into sub modules:
    consistency_checks
    feasibility_fixes
    move the export stuff to dbio

    then there are also convineience functions that need to be ordered at least
    within this file
"""
import logging
from six.moves import zip
from os.path import exists, expanduser, join, abspath
import numpy as np
import vtool as vt
import utool as ut
import cv2
import wbia.constants as const
from wbia.control import controller_inject
from wbia.other.detectfuncs import (
    general_parse_gt,
    general_get_imageset_gids,
    localizer_parse_pred,
    _resize,
    general_overlap,
)
import random

# Inject utool functions
(print, rrr, profile) = ut.inject2(__name__, '[other.detectcore]')
logger = logging.getLogger('wbia')


CLASS_INJECT_KEY, register_ibs_method = controller_inject.make_ibs_register_decorator(
    __name__
)


def nms(dets, scores, thresh, use_cpu=True):
    # Interface into Faster R-CNN's Python native NMS algorithm by Girshick et al.
    from wbia.algo.detect.nms.py_cpu_nms import py_cpu_nms

    return py_cpu_nms(dets, scores, thresh)


@register_ibs_method
def export_to_pascal(ibs, *args, **kwargs):
    """Alias for export_to_xml"""
    return ibs.export_to_xml(*args, **kwargs)


@register_ibs_method
def export_to_xml(
    ibs,
    species_list,
    species_mapping=None,
    offset='auto',
    enforce_viewpoint=False,
    target_size=900,
    purge=False,
    use_maximum_linear_dimension=True,
    use_existing_train_test=True,
    include_parts=False,
    gid_list=None,
    output_path=None,
    allow_empty_images=False,
    min_annot_size=5,
    **kwargs,
):
    """Create training XML for training models."""
    import random
    from datetime import date
    from wbia.detecttools.pypascalmarkup import PascalVOC_Markup_Annotation

    logger.info('Received species_mapping = %r' % (species_mapping,))

    if species_list is None:
        species_list = sorted(set(species_mapping.values()))

    logger.info('Using species_list = %r' % (species_list,))

    def _add_annotation(
        annotation,
        bbox,
        theta,
        species_name,
        viewpoint,
        interest,
        decrease,
        width,
        height,
        part_type=None,
    ):
        # Transformation matrix
        R = vt.rotation_around_bbox_mat3x3(theta, bbox)
        # Get verticies of the annotation polygon
        verts = vt.verts_from_bbox(bbox, close=True)
        # Rotate and transform vertices
        xyz_pts = vt.add_homogenous_coordinate(np.array(verts).T)
        trans_pts = vt.remove_homogenous_coordinate(R.dot(xyz_pts))
        new_verts = np.round(trans_pts).astype(np.int).T.tolist()
        x_points = [pt[0] for pt in new_verts]
        y_points = [pt[1] for pt in new_verts]
        xmin = int(min(x_points) * decrease)
        xmax = int(max(x_points) * decrease)
        ymin = int(min(y_points) * decrease)
        ymax = int(max(y_points) * decrease)
        # Bounds check
        xmin = max(xmin, 0)
        ymin = max(ymin, 0)
        xmax = min(xmax, width - 1)
        ymax = min(ymax, height - 1)
        # Get info
        info = {}

        w_ = xmax - xmin
        h_ = ymax - ymin
        if w_ < min_annot_size:
            return
        if h_ < min_annot_size:
            return

        if viewpoint != -1 and viewpoint is not None:
            info['pose'] = viewpoint

        if interest is not None:
            info['interest'] = '1' if interest else '0'

        if part_type is not None:
            species_name = '%s+%s' % (species_name, part_type)

        area = w_ * h_
        logger.info('\t\tAdding %r with area %0.04f pixels^2' % (species_name, area))

        annotation.add_object(species_name, (xmax, xmin, ymax, ymin), **info)

    current_year = int(date.today().year)
    information = {'database_name': ibs.get_dbname()}
    import datetime

    now = datetime.datetime.now()
    folder = 'VOC%d' % (now.year,)

    if output_path is None:
        output_path = ibs.get_cachedir()

    datadir = join(output_path, 'VOCdevkit', folder)
    imagedir = join(datadir, 'JPEGImages')
    annotdir = join(datadir, 'Annotations')
    setsdir = join(datadir, 'ImageSets')
    mainsetsdir = join(setsdir, 'Main')

    if purge:
        ut.delete(datadir)

    ut.ensuredir(datadir)
    ut.ensuredir(imagedir)
    ut.ensuredir(annotdir)
    ut.ensuredir(setsdir)
    ut.ensuredir(mainsetsdir)

    # Get all gids and process them
    if gid_list is None:
        gid_list = sorted(ibs.get_valid_gids())

    sets_dict = {
        'test': [],
        'train': [],
        'trainval': [],
        'val': [],
    }
    index = 1 if offset == 'auto' else offset

    # Make a preliminary train / test split as imagesets or use the existing ones
    if not use_existing_train_test:
        ibs.imageset_train_test_split(**kwargs)

    train_gid_set = set(general_get_imageset_gids(ibs, 'TRAIN_SET', **kwargs))
    test_gid_set = set(general_get_imageset_gids(ibs, 'TEST_SET', **kwargs))

    logger.info('Exporting %d images' % (len(gid_list),))
    for gid in gid_list:
        aid_list = ibs.get_image_aids(gid)
        image_uri = ibs.get_image_uris(gid)
        image_path = ibs.get_image_paths(gid)
        if len(aid_list) > 0 or allow_empty_images:
            fulldir = image_path.split('/')
            filename = fulldir.pop()
            extension = filename.split('.')[-1]  # NOQA
            out_name = '%d_%06d' % (current_year, index)
            out_img = '%s.jpg' % (out_name,)

            _image = ibs.get_images(gid)
            height, width, channels = _image.shape

            condition = width > height if use_maximum_linear_dimension else width < height
            if condition:
                ratio = height / width
                decrease = target_size / width
                width = target_size
                height = int(target_size * ratio)
            else:
                ratio = width / height
                decrease = target_size / height
                height = target_size
                width = int(target_size * ratio)

            dst_img = join(imagedir, out_img)
            _image = vt.resize(_image, (width, height))
            vt.imwrite(dst_img, _image)

            annotation = PascalVOC_Markup_Annotation(
                dst_img, folder, out_img, source=image_uri, **information
            )
            bbox_list = ibs.get_annot_bboxes(aid_list)
            theta_list = ibs.get_annot_thetas(aid_list)
            species_name_list = ibs.get_annot_species_texts(aid_list)
            viewpoint_list = ibs.get_annot_viewpoints(aid_list)
            interest_list = ibs.get_annot_interest(aid_list)

            part_rowids_list = ibs.get_annot_part_rowids(aid_list)
            zipped = zip(
                bbox_list,
                theta_list,
                species_name_list,
                viewpoint_list,
                interest_list,
                part_rowids_list,
            )
            for (
                bbox,
                theta,
                species_name,
                viewpoint,
                interest,
                part_rowid_list,
            ) in zipped:
                if species_mapping is not None:
                    species_name = species_mapping.get(species_name, species_name)

                if species_name is not None and species_name not in species_list:
                    continue

                _add_annotation(
                    annotation,
                    bbox,
                    theta,
                    species_name,
                    viewpoint,
                    interest,
                    decrease,
                    width,
                    height,
                )

                if include_parts and len(part_rowid_list) > 0:
                    part_bbox_list = ibs.get_part_bboxes(part_rowid_list)
                    part_theta_list = ibs.get_part_thetas(part_rowid_list)
                    part_type_list = ibs.get_part_types(part_rowid_list)
                    part_zipped = zip(part_bbox_list, part_theta_list, part_type_list)
                    for part_bbox, part_theta, part_type in part_zipped:
                        part_viewpoint = viewpoint
                        part_interest = None
                        _add_annotation(
                            annotation,
                            part_bbox,
                            part_theta,
                            species_name,
                            part_viewpoint,
                            part_interest,
                            decrease,
                            width,
                            height,
                            part_type=part_type,
                        )

            out_filename = '%s.xml' % (out_name,)
            dst_annot = join(annotdir, out_filename)

            if gid in test_gid_set:
                sets_dict['test'].append(out_name)
            elif gid in train_gid_set:
                state = random.uniform(0.0, 1.0)
                if state <= 0.75:
                    sets_dict['train'].append(out_name)
                    sets_dict['trainval'].append(out_name)
                else:
                    sets_dict['val'].append(out_name)
                    sets_dict['trainval'].append(out_name)
            else:
                raise AssertionError(
                    'All gids must be either in the TRAIN_SET or TEST_SET imagesets'
                )

            # Write XML
            logger.info(
                'Copying:\n%r\n%r\n%r\n\n' % (image_path, dst_img, (width, height))
            )
            xml_data = open(dst_annot, 'w')
            xml_data.write(annotation.xml())
            xml_data.close()
            while exists(dst_annot):
                index += 1
                if offset != 'auto':
                    break
                out_filename = '%d_%06d.xml' % (current_year, index)
                dst_annot = join(annotdir, out_filename)
        else:
            logger.info('Skipping:\n%r\n\n' % (image_path,))

    for key in sets_dict.keys():
        manifest_filename = '%s.txt' % (key,)
        manifest_filepath = join(mainsetsdir, manifest_filename)
        with open(manifest_filepath, 'w') as file_:
            sets_dict[key].append('')
            content = sets_dict[key]
            content = '\n'.join(content)
            file_.write(content)

    logger.info('...completed')
    return datadir


@register_ibs_method
def export_to_coco(
    ibs,
    species_list,
    species_mapping={},
    viewpoint_mapping={},
    target_size=2400,
    use_maximum_linear_dimension=True,
    use_existing_train_test=True,
    include_parts=False,
    gid_list=None,
    include_reviews=True,
    require_image_reviewed=False,
    require_named=False,
    output_images=True,
    use_global_train_set=False,
    **kwargs,
):
    """Create training COCO dataset for training models."""
    from datetime import date
    import datetime
    import random
    import json

    logger.info('Received species_mapping = %r' % (species_mapping,))
    logger.info('Received viewpoint_mapping = %r' % (viewpoint_mapping,))

    if species_list is None:
        species_list = sorted(set(species_mapping.values()))

    logger.info('Using species_list = %r' % (species_list,))

    current_year = int(date.today().year)
    datadir = abspath(join(ibs.get_cachedir(), 'coco'))
    annotdir = join(datadir, 'annotations')
    imagedir = join(datadir, 'images')
    image_dir_dict = {
        'train': join(imagedir, 'train%s' % (current_year,)),
        'val': join(imagedir, 'val%s' % (current_year,)),
        'test': join(imagedir, 'test%s' % (current_year,)),
    }

    ut.delete(datadir)
    ut.ensuredir(datadir)
    ut.ensuredir(annotdir)
    ut.ensuredir(imagedir)
    for dataset in image_dir_dict:
        ut.ensuredir(image_dir_dict[dataset])

    info = {
        'description': 'Wild Me %s Dataset' % (ibs.dbname,),
        # 'url'          : 'http://www.greatgrevysrally.com',
        'url': 'http://www.wildme.org',
        'version': '1.0',
        'year': current_year,
        'contributor': 'Wild Me <dev@wildme.org>',
        'date_created': datetime.datetime.utcnow().isoformat(' '),
        'name': ibs.get_db_name(),
        'uuid': str(ibs.get_db_init_uuid()),
    }

    licenses = [
        {
            'url': 'http://creativecommons.org/licenses/by-nc-nd/2.0/',
            'id': 3,
            'name': 'Attribution-NonCommercial-NoDerivs License',
        },
    ]

    assert len(species_list) == len(
        set(species_list)
    ), 'Cannot have duplicate species in species_list'
    category_dict = {}
    categories = []
    for index, species in enumerate(sorted(species_list)):

        species = species_mapping.get(species, species)

        categories.append({'id': index, 'name': species, 'supercategory': 'animal'})
        category_dict[species] = index

    def _add_annotation_or_part(
        image_index,
        annot_index,
        annot_uuid,
        bbox,
        theta,
        species_name,
        viewpoint,
        interest,
        annot_name,
        decrease,
        width,
        height,
        individuals,
        part_index=None,
        part_uuid=None,
    ):
        is_part = part_index is not None

        R = vt.rotation_around_bbox_mat3x3(theta, bbox)
        verts = vt.verts_from_bbox(bbox, close=True)
        xyz_pts = vt.add_homogenous_coordinate(np.array(verts).T)
        trans_pts = vt.remove_homogenous_coordinate(R.dot(xyz_pts))
        new_verts = np.round(trans_pts).astype(np.int).T.tolist()

        x_points = [int(np.around(pt[0] * decrease)) for pt in new_verts]
        y_points = [int(np.around(pt[1] * decrease)) for pt in new_verts]
        segmentation = ut.flatten(list(zip(x_points, y_points)))

        xmin = max(min(x_points), 0)
        ymin = max(min(y_points), 0)
        xmax = min(max(x_points), width - 1)
        ymax = min(max(y_points), height - 1)

        w = xmax - xmin
        h = ymax - ymin
        area = w * h

        xtl_, ytl_, w_, h_ = bbox
        xtl_ *= decrease
        ytl_ *= decrease
        w_ *= decrease
        h_ *= decrease

        annot_part = {
            'bbox': [xtl_, ytl_, w_, h_],
            'theta': theta,
            'viewpoint': viewpoint,
            'segmentation': [segmentation],
            'segmentation_bbox': [xmin, ymin, w, h],
            'area': area,
            'iscrowd': 0,
            'id': part_index if is_part else annot_index,
            'image_id': image_index,
            'category_id': category_dict[species_name],
            'uuid': str(part_uuid if is_part else annot_uuid),
            'individual_ids': individuals,
        }
        if is_part:
            annot_part['annot_id'] = annot_index
        else:
            annot_part['isinterest'] = int(interest)
            annot_part['name'] = annot_name

        return annot_part, area

    output_dict = {}
    for dataset in ['train', 'val', 'test']:
        output_dict[dataset] = {
            'info': info,
            'licenses': licenses,
            'categories': categories,
            'images': [],
            'annotations': [],
            'parts': [],
        }

    # Get all gids and process them
    if gid_list is None:
        if require_named:
            aid_list = ibs.get_valid_aids()
            species_list_ = ibs.get_annot_species(aid_list)
            flag_list = [
                species_mapping.get(species_, species_) in species_list
                for species_ in species_list_
            ]
            aid_list = ut.compress(aid_list, flag_list)
            nid_list = ibs.get_annot_nids(aid_list)
            flag_list = [nid >= 0 for nid in nid_list]
            aid_list = ut.compress(aid_list, flag_list)
            gid_list = list(set(ibs.get_annot_gids(aid_list)))
        else:
            gid_list = ibs.get_valid_gids()

        if require_image_reviewed:
            image_reviewed_list = ibs.get_image_reviewed(gid_list)
            gid_list = ut.compress(gid_list, image_reviewed_list)

        gid_list = sorted(list(set(gid_list)))

    # Make a preliminary train / test split as imagesets or use the existing ones
    if not use_existing_train_test:
        ibs.imageset_train_test_split(**kwargs)

    train_gid_set = set(general_get_imageset_gids(ibs, 'TRAIN_SET', **kwargs))
    test_gid_set = set(general_get_imageset_gids(ibs, 'TEST_SET', **kwargs))

    image_index = 1
    annot_index = 1
    part_index = 1

    aid_dict = {}

    logger.info('Exporting %d images' % (len(gid_list),))
    for gid in gid_list:

        if use_global_train_set:
            dataset = 'train'
        else:
            if gid in test_gid_set:
                dataset = 'test'
            elif gid in train_gid_set:
                state = random.uniform(0.0, 1.0)
                if state <= 0.75:
                    dataset = 'train'
                else:
                    dataset = 'val'
            else:
                # raise AssertionError('All gids must be either in the TRAIN_SET or TEST_SET imagesets')
                logger.info('GID = %r was not in the TRAIN_SET or TEST_SET' % (gid,))
                dataset = 'test'

        width, height = ibs.get_image_sizes(gid)
        if target_size is None:
            decrease = 1.0
        else:
            condition = width > height if use_maximum_linear_dimension else width < height
            if condition:
                ratio = height / width
                decrease = target_size / width
                width = target_size
                height = int(target_size * ratio)
            else:
                ratio = width / height
                decrease = target_size / height
                height = target_size
                width = int(target_size * ratio)

        image_path = ibs.get_image_paths(gid)
        image_filename = '%012d.jpg' % (image_index,)
        image_filepath = join(image_dir_dict[dataset], image_filename)

        if output_images:
            _image = ibs.get_images(gid)
            _image = vt.resize(_image, (width, height))
            vt.imwrite(image_filepath, _image)

        image_gps = ibs.get_image_gps(gid)
        if image_gps is None or len(image_gps) != 2 or None in image_gps:
            image_gps_lat, image_gps_lon = None
        else:
            image_gps_lat, image_gps_lon = image_gps
            image_gps_lat = '%03.06f' % (image_gps_lat,)
            image_gps_lon = '%03.06f' % (image_gps_lon,)

        output_dict[dataset]['images'].append(
            {
                'license': 3,
                'file_name': image_filename,
                # 'file_name'     : basename(ibs.get_image_uris_original(gid)),
                'photographer': ibs.get_image_notes(gid),
                'coco_url': None,
                'height': height,
                'width': width,
                'date_captured': ibs.get_image_datetime_str(gid).replace('/', '-'),
                'gps_lat_captured': image_gps_lat,
                'gps_lon_captured': image_gps_lon,
                'flickr_url': None,
                'id': image_index,
                'uuid': str(ibs.get_image_uuids(gid)),
            }
        )

        logger.info(
            'Copying:\n%r\n%r\n%r\n\n' % (image_path, image_filepath, (width, height))
        )

        aid_list = ibs.get_image_aids(gid)
        bbox_list = ibs.get_annot_bboxes(aid_list)
        theta_list = ibs.get_annot_thetas(aid_list)
        species_name_list = ibs.get_annot_species_texts(aid_list)
        viewpoint_list = ibs.get_annot_viewpoints(aid_list)
        interest_list = ibs.get_annot_interest(aid_list)
        annot_uuid_list = ibs.get_annot_uuids(aid_list)
        annot_name_list = ibs.get_annot_name_texts(aid_list)
        part_rowids_list = ibs.get_annot_part_rowids(aid_list)
        nid_list = ibs.get_annot_nids(aid_list)

        zipped = zip(
            aid_list,
            bbox_list,
            theta_list,
            species_name_list,
            viewpoint_list,
            interest_list,
            annot_uuid_list,
            annot_name_list,
            part_rowids_list,
            nid_list,
        )
        for (
            aid,
            bbox,
            theta,
            species_name,
            viewpoint,
            interest,
            annot_uuid,
            annot_name,
            part_rowid_list,
            nid,
        ) in zipped:
            species_name = species_mapping.get(species_name, species_name)

            if species_name is None:
                continue

            if species_name not in species_list:
                continue

            if require_named and nid < 0:
                continue

            viewpoint = viewpoint_mapping.get(species_name, {}).get(viewpoint, viewpoint)

            # if viewpoint is None:
            #     continue

            individuals = ibs.get_name_aids(ibs.get_annot_nids(aid))

            # Transformation matrix
            annot, area = _add_annotation_or_part(
                image_index,
                annot_index,
                annot_uuid,
                bbox,
                theta,
                species_name,
                viewpoint,
                interest,
                annot_name,
                decrease,
                width,
                height,
                individuals,
            )
            logger.info(
                '\t\tAdding annot %r with area %0.04f pixels^2' % (species_name, area)
            )

            if include_reviews:
                reviews = ibs.get_review_rowids_from_single([aid])[0]
                user_list = ibs.get_review_identity(reviews)
                aid_tuple_list = ibs.get_review_aid_tuple(reviews)
                decision_list = ibs.get_review_decision_str(reviews)

                ids = []
                decisions = []
                zipped = zip(user_list, aid_tuple_list, decision_list)
                for user, aid_tuple, decision in zipped:
                    if 'user:web' not in user:
                        continue
                    match = list(set(aid_tuple) - set([aid]))
                    assert len(match) == 1
                    ids.append(match[0])
                    decisions.append(decision.lower())
                annot['review_ids'] = list(zip(ids, decisions))

            output_dict[dataset]['annotations'].append(annot)

            if include_parts and len(part_rowid_list) > 0:

                part_uuid_list = ibs.get_part_uuids(part_rowid_list)
                part_bbox_list = ibs.get_part_bboxes(part_rowid_list)
                part_theta_list = ibs.get_part_thetas(part_rowid_list)
                part_type_list = ibs.get_part_types(part_rowid_list)

                part_zipped = zip(
                    part_uuid_list, part_bbox_list, part_theta_list, part_type_list
                )
                for part_uuid, part_bbox, part_theta, part_type in part_zipped:
                    part_species_name = '%s+%s' % (species_name, part_type)

                    part_species_name = species_mapping.get(
                        part_species_name, part_species_name
                    )

                    if part_species_name is None:
                        continue

                    if part_species_name not in species_list:
                        continue

                    part, area = _add_annotation_or_part(
                        image_index,
                        annot_index,
                        annot_uuid,
                        part_bbox,
                        part_theta,
                        part_species_name,
                        viewpoint,
                        interest,
                        annot_name,
                        decrease,
                        width,
                        height,
                        individuals,
                        part_index=part_index,
                        part_uuid=part_uuid,
                    )
                    logger.info(
                        '\t\tAdding part %r with area %0.04f pixels^2'
                        % (part_species_name, area)
                    )
                    output_dict[dataset]['parts'].append(part)

                part_index += 1

            aid_dict[aid] = annot_index
            annot_index += 1

        image_index += 1

    for dataset in output_dict:
        annots = output_dict[dataset]['annotations']
        for index in range(len(annots)):
            annot = annots[index]

            # Map internal aids to external annot index
            individual_ids = annot['individual_ids']
            individual_ids_ = []
            for individual_id in individual_ids:
                if individual_id not in aid_dict:
                    continue
                individual_id_ = aid_dict[individual_id]
                individual_ids_.append(individual_id_)
            annot['individual_ids'] = individual_ids_

            # Map reviews
            if include_reviews:
                review_ids = annot['review_ids']
                review_ids_ = []
                for review in review_ids:
                    review_id, review_decision = review
                    if review_id not in aid_dict:
                        continue
                    review_id_ = aid_dict[review_id]
                    review_ = (
                        review_id_,
                        review_decision,
                    )
                    review_ids_.append(review_)
                annot['review_ids'] = review_ids_

            # Store
            output_dict[dataset]['annotations'][index] = annot

    for dataset in output_dict:
        json_filename = 'instances_%s%s.json' % (dataset, current_year)
        json_filepath = join(annotdir, json_filename)

        with open(json_filepath, 'w') as json_file:
            json.dump(output_dict[dataset], json_file)

    logger.info('...completed')
    return datadir


@register_ibs_method
def imageset_train_test_split(
    ibs, train_split=0.8, is_tile=False, gid_list=None, **kwargs
):
    from random import shuffle

    if gid_list is None:
        gid_list = ibs.get_valid_gids(is_tile=is_tile)

    aids_list = ibs.get_image_aids(gid_list)
    distro_dict = {}
    for gid, aid_list in zip(gid_list, aids_list):
        total = len(aid_list)
        if total not in distro_dict:
            distro_dict[total] = []
        distro_dict[total].append(gid)

    logger.info('Processing train/test imagesets...')
    global_train_list = []
    global_test_list = []
    for distro, gid_list_ in distro_dict.items():
        total = len(gid_list_)
        shuffle(gid_list_)
        split_index = total * (1.0 - train_split) + 1e-9  # weird
        if split_index < 1.0:
            split_index = total / 2
        else:
            split_index = np.around(split_index)
        split_index = int(split_index)
        args = (
            distro,
            total,
            split_index,
        )
        logger.info('\tnum aids distro: %d - total: %d - split_index: %d' % args)
        train_list = gid_list_[split_index:]
        test_list = gid_list_[:split_index]
        args = (
            len(test_list),
            len(train_list),
            len(train_list) / total,
        )
        logger.info('\ttest: %d\n\ttrain: %d\n\tsplit: %0.04f' % args)
        global_train_list.extend(train_list)
        global_test_list.extend(test_list)

    args = (
        len(global_train_list),
        len(global_test_list),
        len(global_train_list) + len(global_test_list),
        len(global_train_list) / len(gid_list),
        train_split,
    )

    train_imgsetid = ibs.add_imagesets('TRAIN_SET')
    test_imgsetid = ibs.add_imagesets('TEST_SET')

    temp_list = ibs.get_imageset_gids(train_imgsetid)
    ibs.unrelate_images_and_imagesets(temp_list, [train_imgsetid] * len(temp_list))
    temp_list = ibs.get_imageset_gids(test_imgsetid)
    ibs.unrelate_images_and_imagesets(temp_list, [test_imgsetid] * len(temp_list))

    ibs.set_image_imgsetids(global_train_list, [train_imgsetid] * len(global_train_list))
    ibs.set_image_imgsetids(global_test_list, [test_imgsetid] * len(global_test_list))

    logger.info('Complete... %d train + %d test = %d (%0.04f %0.04f)' % args)


@register_ibs_method
def localizer_distributions(ibs, threshold=10, dataset=None):
    # Process distributions of densities
    if dataset is None:
        gid_list = ibs.get_valid_gids()
    else:
        assert dataset in ['TRAIN_SET', 'TEST_SET']
        imageset_id = ibs.get_imageset_imgsetids_from_text(dataset)
        gid_list = list(set(ibs.get_imageset_gids(imageset_id)))
    aids_list = ibs.get_image_aids(gid_list)
    distro_dict = {}
    species_dict = {}
    for gid, aid_list in zip(gid_list, aids_list):
        total = len(aid_list)
        if total >= threshold:
            total = threshold
        if total not in distro_dict:
            distro_dict[total] = 0
        total = '%s' % (total,) if total < threshold else '%s+' % (total,)
        distro_dict[total] += 1
        for aid in aid_list:
            species = ibs.get_annot_species_texts(aid)
            viewpoint = ibs.get_annot_viewpoints(aid)
            if species not in species_dict:
                species_dict[species] = {}
            if viewpoint not in species_dict[species]:
                species_dict[species][viewpoint] = 0
            species_dict[species][viewpoint] += 1

    logger.info('Annotation density distribution (annotations per image)')
    for distro in sorted(distro_dict.keys()):
        logger.info('{:>6} annot(s): {:>5}'.format(distro, distro_dict[distro]))
    logger.info('')

    for species in sorted(species_dict.keys()):
        logger.info('Species viewpoint distribution: %r' % (species,))
        viewpoint_dict = species_dict[species]
        total = 0
        for viewpoint in const.VIEWTEXT_TO_VIEWPOINT_RADIANS:
            count = viewpoint_dict.get(viewpoint, 0)
            logger.info('{:>15}: {:>5}'.format(viewpoint, count))
            total += count
        logger.info('TOTAL: %d\n' % (total,))

    # visualize_distributions(distro_dict, threshold=threshold)


def visualize_distributions(distro_dict, threshold=10):
    import matplotlib.pyplot as plt

    key_list = sorted(distro_dict.keys())
    threshold_str = '%d+' % (threshold,)
    label_list = [threshold_str if key == threshold else str(key) for key in key_list]
    size_list = [distro_dict[key] for key in key_list]
    color_list = ['yellowgreen', 'gold', 'lightskyblue', 'lightcoral']
    explode = [0.0] + [0.0] * (len(size_list) - 1)

    plt.pie(
        size_list,
        explode=explode,
        labels=label_list,
        colors=color_list,
        autopct='%1.1f%%',
        shadow=True,
        startangle=90,
    )
    plt.axis('equal')
    plt.show()


@register_ibs_method
def visualize_pascal_voc_dataset(
    ibs, dataset_path, num_examples=30, randomize=False, write=True, write_path=None
):
    r"""Visualize the PASCAL VOC dataset.

    Args:
        ibs (IBEISController):
        dataset_path (str): the dataset path in the PASCAL VOC format
        num_examples (int, optional): the number of examples to draw
        randomize (bool, optional): if to randomize the visualization
        write (bool, optional): if to display or write the files

    CommandLine:
        python -m wbia.other.detectcore --test-visualize_pascal_voc_dataset

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.other.detectcore import *  # NOQA
        >>> import wbia  # NOQA
        >>> ibs = wbia.opendb('testdb1')
        >>> dataset_path = '/Users/jason.parham/Downloads/wilddog_data/data/VOCdevkit/VOC2020/'
        >>> # dataset_path = '/Users/jason.parham/Downloads/LearningData/'
        >>> # dataset_path = '/Users/jason.parham/Downloads/VOCdevkit/VOC2018/'
        >>> ibs.visualize_pascal_voc_dataset(dataset_path, randomize=True)
    """
    from wbia.detecttools.wbiadata import IBEIS_Data
    import random

    num_examples = int(num_examples)
    assert num_examples > 0

    dataset = IBEIS_Data(dataset_path)
    dataset.print_distribution()

    image_list = sorted(dataset.images)

    num_examples = min(num_examples, len(image_list))

    if randomize:
        random.shuffle(image_list)

    if write_path is None:
        write_path = abspath(expanduser(join('~', 'Desktop')))

    for image in image_list[:num_examples]:
        if write:
            write_filepath = join(write_path, image.filename)
            image = image.show(display=False)
            cv2.imwrite(write_filepath, image)
        else:
            image.show()


@register_ibs_method
def classifier_visualize_training_localizations(
    ibs,
    classifier_weight_filepath,
    species_list=['zebra'],
    scheme=2,
    output_path=None,
    values=None,
    **kwargs,
):
    def _draw(image_dict, list_, color):
        for _ in list_:
            vals = _['gid'], _['xbr'], _['ybr'], _['xtl'], _['ytl']
            gid, xbr, ybr, xtl, ytl = vals
            height, width = image_dict[gid].shape[:2]
            xbr = int(xbr * width)
            ybr = int(ybr * height)
            xtl = int(xtl * width)
            ytl = int(ytl * height)
            image = image_dict[gid]
            cv2.rectangle(image, (xtl, ytl), (xbr, ybr), color, 4)

    def _write_chips(chip_list, output_path_fmt_str):
        interpolation = cv2.INTER_LANCZOS4
        warpkw = dict(interpolation=interpolation)
        chip_list = []
        for _ in list_:
            vals = _['gid'], _['xbr'], _['ybr'], _['xtl'], _['ytl']
            gid, xbr, ybr, xtl, ytl = vals
            height, width = image_dict[gid].shape[:2]
            xbr = int(xbr * width)
            ybr = int(ybr * height)
            xtl = int(xtl * width)
            ytl = int(ytl * height)
            image = image_dict[gid]
            # Get chips
            chip = image[ytl:ybr, xtl:xbr, :]
            chip = cv2.resize(chip, (192, 192), **warpkw)
            chip_list.append(chip)
        return chip_list

    # Get output path
    if output_path is None:
        output_path = abspath(expanduser(join('~', 'Desktop', 'output-bootstrap')))
    ut.delete(output_path)
    ut.ensuredir(output_path)

    if values is None:
        # Load data
        logger.info('Loading pre-trained features for filtered localizations')
        train_gid_list = general_get_imageset_gids(ibs, 'TRAIN_SET', **kwargs)
        train_gid_list = train_gid_list[:10]

        config = {
            'algo': '_COMBINED',
            'species_set': set(species_list),
            'features': True,
            'feature2_algo': 'resnet',
            'classify': True,
            'classifier_algo': 'svm',
            'classifier_weight_filepath': classifier_weight_filepath,
            'nms': True,
            'nms_thresh': 0.50,
            # 'thresh'       : True,
            # 'index_thresh' : 0.25,
        }

        logger.info('\tGather Ground-Truth')
        gt_dict = general_parse_gt(ibs, test_gid_list=train_gid_list, **config)

        logger.info('\tGather Predictions')
        pred_dict = localizer_parse_pred(ibs, test_gid_list=train_gid_list, **config)

        logger.info('Mine proposals')
        reviewed_gid_dict = {}
        values = _bootstrap_mine(
            ibs, gt_dict, pred_dict, scheme, reviewed_gid_dict, **kwargs
        )

    mined_gid_list, mined_gt_list, mined_pos_list, mined_neg_list = values

    logger.info('Prepare images')
    # Get images and a dictionary based on their gids
    image_list = ibs.get_images(mined_gid_list)
    image_dict = {gid: image for gid, image in zip(mined_gid_list, image_list)}

    # Draw positives
    list_ = mined_pos_list
    color = (0, 255, 0)
    chip_list = _draw(image_dict, list_, color)
    pos_path = join(output_path, 'positives')
    ut.ensuredir(pos_path)
    _write_chips(chip_list, join(pos_path, 'chips_pos_%05d.png'))

    # Draw negatives
    list_ = mined_neg_list
    color = (0, 0, 255)
    chip_list = _draw(image_dict, list_, color)
    neg_path = join(output_path, 'negatives')
    ut.ensuredir(neg_path)
    _write_chips(chip_list, join(neg_path, 'chips_neg_%05d.png'))

    # Draw positives
    list_ = mined_gt_list
    color = (255, 0, 0)
    _draw(image_dict, list_, color)

    logger.info('Write images to %r' % (output_path,))
    # Write images to disk
    for gid in image_dict:
        output_filename = 'localizations_gid_%d.png' % (gid,)
        output_filepath = join(output_path, output_filename)
        cv2.imwrite(output_filepath, image_dict[gid])


@register_ibs_method
def redownload_detection_models(ibs):
    r"""Re-download detection models.

    Args:
        ibs (IBEISController):

    CommandLine:
        python -c "from wbia.algo.detect import grabmodels; grabmodels.redownload_models()"
        python -c "import utool, wbia.algo; utool.view_directory(wbia.algo.detect.grabmodels._expand_modeldir())"

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.other.detectcore import *  # NOQA
        >>> import wbia  # NOQA
        >>> ibs = wbia.opendb('testdb1')
        >>> result = redownload_detection_models(ibs)
        >>> print(result)
    """
    logger.info('[other.detectcore] redownload_detection_models')
    from wbia.algo.detect import grabmodels

    modeldir = ibs.get_detect_modeldir()
    grabmodels.redownload_models(modeldir=modeldir)


@register_ibs_method
def view_model_dir(ibs):
    logger.info('[other.detectcore] redownload_detection_models')
    modeldir = ibs.get_detect_modeldir()
    ut.view_directory(modeldir)
    # grabmodels.redownload_models(modeldir=modeldir)


def _bootstrap_mine(
    ibs,
    gt_dict,
    pred_dict,
    scheme,
    reviewed_gid_dict,
    min_overlap=0.75,
    max_overlap=0.25,
):
    import random

    ##################################################################################
    # Step 7.5: gather SVM training data from overlap images
    #           note that this step randomly subsamples new negatives, so
    #           each SVM in the ensemble is given a different set of negatives
    mined_gid_list = []
    mined_gt_list = []
    mined_pos_list = []
    mined_neg_list = []
    for image_uuid in gt_dict:
        # logger.info('--- Processing user interaction for image %r' % (image_uuid, ))
        # Get the gt and prediction list
        gt_list = gt_dict[image_uuid]
        pred_list = pred_dict[image_uuid]

        # If never seen this image before, pick a new selection of GT bboxes
        image_gid = ibs.get_image_gids_from_uuid(image_uuid)
        if image_gid not in reviewed_gid_dict:
            # Simulate the user selecting the gt bounding box(es)
            index_list = list(range(len(gt_list)))
            if scheme == 1:
                # Pick a random bbox
                index_list_ = [random.choice(index_list)]
            elif scheme == 2:
                # Pick all gt boxes
                index_list_ = index_list[:]
            else:
                raise ValueError
            reviewed_gid_dict[image_gid] = index_list_

        # Filter based on picked bboxes for gt
        picked_index_list = reviewed_gid_dict[image_gid]
        gt_list_ = [gt_list[picked_index] for picked_index in picked_index_list]

        # Calculate overlap
        overlap = general_overlap(gt_list_, pred_list)
        num_gt, num_pred = overlap.shape

        if num_gt == 0 or num_pred == 0:
            continue
        else:
            pos_idx_list = np.where(overlap >= min_overlap)[1]
            neg_idx_list = np.where(overlap <= max_overlap)[1]

            num_pos = len(pos_idx_list)
            num_neg = len(neg_idx_list)

            # Randomly sample negative chips to get new candidates
            # Most of the time (like almost always will happen)
            if num_neg > num_pos:
                np.random.shuffle(neg_idx_list)
                neg_idx_list = neg_idx_list[:num_pos]

            mined_gid_list.append(image_gid)
            mined_gt_list += gt_list_
            mined_pos_list += [pred_list[idx] for idx in pos_idx_list]
            mined_neg_list += [pred_list[idx] for idx in neg_idx_list]

    args = (
        len(mined_pos_list),
        len(mined_neg_list),
        len(mined_gid_list),
    )
    logger.info('Mined %d positive, %d negative from %d images' % args)

    return mined_gid_list, mined_gt_list, mined_pos_list, mined_neg_list


@register_ibs_method
def visualize_ground_truth(ibs, config, **kwargs):
    visualize_bounding_boxes(ibs, config, 'ground_truth', **kwargs)


@register_ibs_method
def visualize_predictions(ibs, config, **kwargs):
    visualize_bounding_boxes(ibs, config, 'prediction', **kwargs)


def visualize_bounding_boxes(
    ibs,
    config,
    version,
    gid_list=None,
    randomize=False,
    num_images=10,
    t_width=500,
    output_path=None,
):
    if gid_list is None:
        gid_list = general_get_imageset_gids(ibs, 'TEST_SET', **config)
    else:
        num_images = None

    if randomize:
        random.shuffle(gid_list)

    if num_images not in [-1, None]:
        num_images = min(num_images, len(gid_list))
        gid_list = gid_list[:num_images]

    uuid_list = ibs.get_image_uuids(gid_list)

    assert version is not None
    version = version.lower()
    if version == 'prediction':
        logger.info('\tGather Predictions')
        val_dict = localizer_parse_pred(ibs, test_gid_list=gid_list, **config)
    elif version == 'ground_truth':
        logger.info('\tGather Ground-Truth')
        val_dict = general_parse_gt(ibs, test_gid_list=gid_list, **config)

    if output_path is None:
        output_path = abspath(expanduser(join('~', 'Desktop', 'bboxes')))
        ut.ensuredir(output_path)

    filepath_dict = {}
    for gid, image_uuid in zip(gid_list, uuid_list):
        image = ibs.get_images(gid)
        image = _resize(image, t_width=t_width)
        h, w, c = image.shape

        val_list = val_dict[image_uuid]

        for val in val_list:
            xbr = int(np.around(val['xbr'] * w))
            ybr = int(np.around(val['ybr'] * h))
            xtl = int(np.around(val['xtl'] * w))
            ytl = int(np.around(val['ytl'] * h))
            cv2.rectangle(image, (xtl, ytl), (xbr, ybr), (0, 140, 255), 4)

        write_filename = 'bboxes_%d_%s.png' % (gid, version)
        write_filepath = join(output_path, write_filename)
        logger.info(write_filepath)
        cv2.imwrite(write_filepath, image)

        filepath_dict[gid] = write_filepath

    return filepath_dict



# -*- coding: utf-8 -*-
# Autogenerated on 12:54:39 2017/06/25
# flake8: noqa
import logging
from wbia.algo.verif import clf_helpers
from wbia.algo.verif import sklearn_utils
from wbia.algo.verif import deploy
from wbia.algo.verif import pairfeat
from wbia.algo.verif import verifier
from wbia.algo.verif import vsone
import utool

print, rrr, profile = utool.inject2(__name__, '[wbia.algo.verif]')
logger = logging.getLogger('wbia')


def reassign_submodule_attributes(verbose=1):
    """
    Updates attributes in the __init__ modules with updated attributes
    in the submodules.
    """
    import sys

    if verbose and '--quiet' not in sys.argv:
        print('dev reimport')
    # Self import
    import wbia.algo.verif

    # Implicit reassignment.
    seen_ = set([])
    for tup in IMPORT_TUPLES:
        if len(tup) > 2 and tup[2]:
            continue  # dont import package names
        submodname, fromimports = tup[0:2]
        submod = getattr(wbia.algo.verif, submodname)
        for attr in dir(submod):
            if attr.startswith('_'):
                continue
            if attr in seen_:
                # This just holds off bad behavior
                # but it does mimic normal util_import behavior
                # which is good
                continue
            seen_.add(attr)
            setattr(wbia.algo.verif, attr, getattr(submod, attr))


def reload_subs(verbose=1):
    """ Reloads wbia.algo.verif and submodules """
    if verbose:
        print('Reloading wbia.algo.verif submodules')
    rrr(verbose > 1)

    def wrap_fbrrr(mod):
        def fbrrr(*args, **kwargs):
            """ fallback reload """
            if verbose > 0:
                print('Auto-reload (using rrr) not setup for mod=%r' % (mod,))

        return fbrrr

    def get_rrr(mod):
        if hasattr(mod, 'rrr'):
            return mod.rrr
        else:
            return wrap_fbrrr(mod)

    def get_reload_subs(mod):
        return getattr(mod, 'reload_subs', wrap_fbrrr(mod))

    get_rrr(clf_helpers)(verbose > 1)
    get_rrr(sklearn_utils)(verbose > 1)
    get_rrr(vsone)(verbose > 1)
    rrr(verbose > 1)
    try:
        # hackish way of propogating up the new reloaded submodule attributes
        reassign_submodule_attributes(verbose=verbose)
    except Exception as ex:
        print(ex)


rrrr = reload_subs

IMPORT_TUPLES = [
    ('clf_helpers', None),
    ('sklearn_utils', None),
    ('vsone', None),
    ('deploy', None),
    ('verifier', None),
    ('pairfeat', None),
]
"""
Regen Command:
    cd /home/joncrall/code/wbia/wbia/algo/verif
    makeinit.py --modname=wbia.algo.verif
"""



# -*- coding: utf-8 -*-
import logging
from os.path import split, splitext, join, exists, realpath
import datetime
import distutils
import utool as ut

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

VERBOSE_SQL = ut.get_argflag(('--print-sql', '--verbose-sql', '--verb-sql', '--verbsql'))
NOT_QUIET = not (ut.QUIET or ut.get_argflag('--quiet-sql'))


def compare_string_versions(a, b):
    r"""
    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control._sql_helpers import *  # NOQA
        >>> a = '1.1.1'
        >>> b = '1.0.0'
        >>> result1 = compare_string_versions(a, b)
        >>> result2 = compare_string_versions(b, a)
        >>> result3 = compare_string_versions(a, a)
        >>> result = ', '.join(map(str, [result1, result2, result3]))
        >>> print(result)
        1, -1, 0
    """
    va = distutils.version.LooseVersion(a)
    vb = distutils.version.LooseVersion(b)
    if va > vb:
        return 1
    elif va < vb:
        return -1
    elif va == vb:
        return 0
    raise AssertionError(
        '[!update_schema_version] Two version numbers are '
        'the same along the update path'
    )


def _devcheck_backups():
    from wbia import dtool as dt

    dbdir = ut.truepath('~/work/PZ_Master1/_ibsdb')
    sorted(ut.glob(join(dbdir, '_wbia_backups'), '*staging_back*.sqlite3'))
    fpaths = sorted(ut.glob(join(dbdir, '_wbia_backups'), '*database_back*.sqlite3'))
    for fpath in fpaths:
        db_uri = 'file://{}'.format(realpath(fpath))
        db = dt.SQLDatabaseController.from_uri(db_uri)
        logger.info('fpath = %r' % (fpath,))
        num_edges = len(db.executeone('SELECT rowid from annotmatch'))
        logger.info('num_edges = %r' % (num_edges,))
        num_names = len(db.executeone('SELECT DISTINCT name_rowid from annotations'))
        logger.info('num_names = %r' % (num_names,))
        # df = db.get_table_as_pandas('annotations', columns=['annot_rowid',
        #                                                     'name_rowid'])


def fix_metadata_consistency(db):
    """
    duct tape function

    db.print_table_csv('metadata')
    """
    db.print_table_csv('metadata')
    metadata_items = db.get_metadata_items()
    new_metadata_items = []
    for key, val in metadata_items:
        # FIX OLDSTYLE SUPERKEYS
        if key.endswith('_constraint'):
            # the constraint metadata keys were all depricated
            continue
        if key.endswith('_superkeys'):
            if val.find(';') > -1:
                newval = repr([tuple(val.split(';'))])
            elif not val.startswith('['):
                newval = repr([(val,)])
            else:
                continue
            new_metadata_items.append((key, newval))
        else:
            logger.info('--')
            logger.info(key)
            logger.info(val)
    for key, val in new_metadata_items:
        db.set_metadata_val(key, val)


# =========================
# Database Backup Functions
# =========================


MAX_KEEP = 2048


def revert_to_backup(ibs):
    r"""
    Args:
        db_dir (?):

    CommandLine:
        python -m wbia.control._sql_helpers --exec-revert_to_backup

    Example:
        >>> # SCRIPT
        >>> from wbia.control._sql_helpers import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='elephants')
        >>> result = revert_to_backup(ibs)
        >>> print(result)
    """
    db_path = ibs.get_db_core_path()
    staging_path = ibs.get_db_staging_path()

    ibs.disconnect_sqldatabase()
    backup_dir = ibs.backupdir

    # Core database
    fname, ext = splitext(db_path)
    db_path_ = '%s_revert.sqlite3' % (fname,)
    ut.move(db_path, db_path_)
    fpath, fname = split(fname)
    path_list = sorted(ut.glob(backup_dir, '%s_*%s' % (fname, ext)))
    assert len(path_list) > 0
    previous_backup = path_list[-1]
    copy_database(previous_backup, db_path)

    # Staging database
    fname, ext = splitext(staging_path)
    staging_path_ = '%s_revert.sqlite3' % (fname,)
    ut.move(staging_path, staging_path_)
    fpath, fname = split(fname)
    path_list = sorted(ut.glob(backup_dir, '%s_*%s' % (fname, ext)))
    assert len(path_list) > 0
    previous_backup = path_list[-1]
    copy_database(previous_backup, staging_path)

    # Delete the cache
    ut.delete(ibs.cachedir)


def ensure_daily_database_backup(db_dir, db_fname, backup_dir, max_keep=MAX_KEEP):
    database_backup(db_dir, db_fname, backup_dir, max_keep=max_keep, manual=False)


def get_backupdir(db_dir, db_fname):
    r"""
    CommandLine:
        python -m _sql_helpers get_backupdir --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control._sql_helpers import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> db_dir = ibs.get_ibsdir()
        >>> db_fname = ibs.sqldb_fname
        >>> backup_dir = ibs.backupdir
        >>> result = get_backupdir(db_dir, db_fname)
    """
    pass


def get_backup_fpaths(ibs):
    fname, ext = splitext(ibs.sqldb_fname)
    backups = sorted(ut.glob(ibs.backupdir, '*%s' % ext))
    # backup_info = [ut.get_file_info(fpath) for fpath in backups]
    modified = [ut.get_file_info(fpath)['last_modified'] for fpath in backups]
    unixtimes = [ut.util_time.exiftime_to_unixtime(tag) for tag in modified]
    backups = ut.sortedby(backups, unixtimes)
    return backups
    # backup_uuids = [ut.get_file_uuid(fpath) for fpath in backups]
    # backup_hashes = [ut.get_file_hash(fpath) for fpath in backups]
    # backup_bytes = [ut.get_file_nBytes(fpath) for fpath in backups]
    pass


def copy_database(src_fpath, dst_fpath):
    from wbia import dtool

    # Load database and ask it to copy itself, which enforces an exclusive
    # blocked lock for all processes potentially writing to the database
    timeout = 12 * 60 * 60  # Allow a lock of up to 12 hours for a database backup routine
    if not src_fpath.startswith('file:'):
        src_fpath = 'file://{}'.format(realpath(src_fpath))
    db = dtool.SQLDatabaseController.from_uri(src_fpath, timeout=timeout)
    db.backup(dst_fpath)


def database_backup(db_dir, db_fname, backup_dir, max_keep=MAX_KEEP, manual=True):
    """
    >>> db_dir = ibs.get_ibsdir()
    >>> db_fname = ibs.sqldb_fname
    >>> backup_dir = ibs.backupdir
    >>> max_keep = MAX_KEEP
    >>> manual = False
    """
    fname, ext = splitext(db_fname)
    src_fpath = join(db_dir, db_fname)
    # now = datetime.datetime.now()
    now = datetime.datetime.utcnow()
    if manual:
        now_str = now.strftime('%Y_%m_%d_%H_%M_%S')
    else:
        now_str = now.strftime('%Y_%m_%d_00_00_00')
    # dst_fpath = join(backup_dir, '%s_backup_%s%s' % (fname, now_str, ext))
    dst_fname = ''.join((fname, '_backup_', now_str, ext))
    dst_fpath = join(backup_dir, dst_fname)
    if exists(src_fpath) and not exists(dst_fpath):
        logger.info(
            '[ensure_daily_database_backup] Daily backup of database: %r -> %r'
            % (src_fpath, dst_fpath)
        )
        copy_database(src_fpath, dst_fpath)
        # Clean-up old database backups
        remove_old_backups(backup_dir, ext, max_keep)


def remove_old_backups(backup_dir, ext, max_keep):
    path_list = sorted(ut.glob(backup_dir, '*%s' % ext))
    if len(path_list) > max_keep:
        path_delete_list = path_list[: -1 * max_keep]
        for path_delete in path_delete_list:
            logger.info(
                '[ensure_daily_database_backup] Deleting old backup %r' % path_delete
            )
            ut.remove_file(path_delete, verbose=False)


# ========================
# Schema Updater Functions
# ========================


@profile
def ensure_correct_version(
    ibs, db, version_expected, schema_spec, dobackup=True, verbose=ut.NOT_QUIET
):
    """
    FIXME: AN SQL HELPER FUNCTION SHOULD BE AGNOSTIC TO CONTROLER OBJECTS

    ensure_correct_version

    Args:
        ibs (IBEISController):
        db (SQLController):
        version_expected (str): version you want to be at
        schema_spec (module): schema module
        dobackup (bool):

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control._sql_helpers import *  # NOQA
        >>> ibs = '?'
        >>> db = ibs.db
        >>> version_expected = ibs.db_version_expected
        >>> schema_spec = DB_SCHEMA
        >>> dobackup = True
        >>> result = ensure_correct_version(ibs, db, version_expected, schema_spec, dobackup)
        >>> print(result)

    Args:
        schema_spec (module): module of schema specifications
    """
    from wbia import constants as const
    from wbia import params

    params.parse_args()

    # logger.info('[SQL_] ensure_correct_version')
    force_incremental = params.args.force_incremental_db_update
    want_base_version = version_expected == const.BASE_DATABASE_VERSION
    if want_base_version:
        logger.info('[SQL_] base version expected... returning')
        # Nothing to do. autogenerated file is pointless
        return
    version = db.get_db_version()
    # NEW DATABASE CONDITION
    is_base_version = version == const.BASE_DATABASE_VERSION
    # <DEBUG>
    # if ut.get_flag('--verbsql') or ut.VERBOSE or True:
    #    key_list = locals().keys()
    #    keystr_list = sorted(ut.parse_locals_keylist(locals(), key_list))
    #    logger.info('KEYLIST:' + ut.indentjoin(keystr_list, '\n * '))
    # </DEBUG>
    # +-----------------------------------
    # SKIP TO CURRENT VERSION IF POSSIBLE
    # +-----------------------------------
    can_skip = is_base_version and not force_incremental
    if can_skip:
        # logger.info('[SQL_] we can skip')
        # Check to see if a prebuilt current schema_spec module exists
        current_schema_exists = (
            schema_spec.UPDATE_CURRENT is not None
            and schema_spec.VERSION_CURRENT is not None
        )
        if current_schema_exists:
            # check to see if more than the metadata table exists
            is_newdb = db.get_table_names() == [const.METADATA_TABLE]
            current_schema_compatible = (
                is_newdb and schema_spec.VERSION_CURRENT <= version_expected
            )
            if current_schema_compatible:
                # Since this is a new database, we do not have to worry about backinng up the
                # current database.  The subsequent update functions (if needed) will handle
                # this for us.
                if verbose:
                    logger.info('[_SQL] New database and a current schema found')
                schema_spec.UPDATE_CURRENT(db, ibs=ibs)
                db.set_db_version(schema_spec.VERSION_CURRENT)
                if verbose:
                    logger.info(
                        '[_SQL] Database version updated (skipped) to %r '
                        % (schema_spec.VERSION_CURRENT)
                    )
            else:
                logger.info(
                    '[_SQL] Current database is not compatible, updating incrementally...'
                )
        else:
            logger.info(
                '[_SQL] New database but current version not exported, updating incrementally...'
            )
    # +--------------------------------------
    # INCREMENTAL UPDATE TO EXPECTED VERSION
    # +--------------------------------------
    # Check version again for sanity's sake, update if exported current is behind expected
    version = db.get_db_version()
    if verbose:
        logger.info(
            '[_SQL.%s] Database version: %r | Expected version: %r '
            % (ut.get_caller_name(), version, version_expected)
        )
    if version < version_expected:
        logger.info('[_SQL] Database version behind, updating...')
        update_schema_version(
            ibs, db, schema_spec, version, version_expected, dobackup=dobackup
        )
        db.set_db_version(version_expected)
        logger.info(
            '[_SQL] Database version updated (incrementally) to %r' % (version_expected)
        )
    elif version > version_expected:
        msg = (
            '[_SQL] ERROR: ' 'Expected database version behind. expected: %r. got: %r'
        ) % (version_expected, version)
        raise AssertionError(msg)


@profile
def update_schema_version(
    ibs, db, schema_spec, version, version_target, dobackup=True, clearbackup=False
):
    """
    version_target = version_expected
    clearbackup = False
    FIXME: AN SQL HELPER FUNCTION SHOULD BE AGNOSTIC TO CONTROLER OBJECTS
    """

    def _check_superkeys():
        all_tablename_list = db.get_table_names()
        # always ignore the metadata table.
        ignore_tables_ = ['metadata']
        tablename_list = [
            tablename
            for tablename in all_tablename_list
            if tablename not in ignore_tables_
        ]
        for tablename in tablename_list:
            superkey_colnames_list = db.get_table_superkey_colnames(tablename)
            # some tables seem to only have old constraints and aren't
            # properly updated to superkeys... weird.
            old_constraints = db.get_table_constraints(tablename)
            assert (
                len(superkey_colnames_list) > 0 or len(old_constraints) > 0
            ), 'ERROR UPDATING DATABASE, SUPERKEYS of %s DROPPED!' % (tablename,)

    logger.info('[_SQL] update_schema_version')
    db_fpath = db.uri.replace('file://', '')
    if dobackup:
        db_dpath, db_fname = split(db_fpath)
        db_fname_noext, ext = splitext(db_fname)
        db_backup_fname = ''.join((db_fname_noext, '_backup', '_v', version, ext))
        db_backup_fpath = join(db_dpath, db_backup_fname)
        count = 0
        # TODO MAKE UTOOL THAT DOES THIS (there might be one in util_logging)
        while ut.checkpath(db_backup_fpath, verbose=True):
            db_backup_fname = ''.join(
                (db_fname_noext, '_backup', '_v', version, '_copy', str(count), ext)
            )
            db_backup_fpath = join(db_dpath, db_backup_fname)
            count += 1
        copy_database(db_fpath, db_backup_fpath)

    legacy_update_funcs = schema_spec.LEGACY_UPDATE_FUNCTIONS
    for legacy_version, func in legacy_update_funcs:
        if compare_string_versions(version, legacy_version) == -1:
            func(db)
    db_versions = schema_spec.VALID_VERSIONS
    import functools

    _key = functools.cmp_to_key(compare_string_versions)
    valid_versions = sorted(db_versions.keys(), key=_key)
    logger.info('valid_versions = %r' % (valid_versions,))
    try:
        start_index = valid_versions.index(version) + 1
    except IndexError:
        raise AssertionError(
            '[!update_schema_version]' ' The current database version is unknown'
        )
    try:
        end_index = valid_versions.index(version_target) + 1
    except IndexError:
        raise AssertionError(
            '[!update_schema_version]' ' The target database version is unknown'
        )

    try:
        logger.info('Update path: %r ' % (valid_versions[start_index:end_index]))
        for index in range(start_index, end_index):
            next_version = valid_versions[index]
            logger.info('Updating database to version: %r' % (next_version))
            pre, update, post = db_versions[next_version]
            if pre is not None:
                pre(db, ibs=ibs)
            if update is not None:
                update(db, ibs=ibs)
            if post is not None:
                post(db, ibs=ibs)
            _check_superkeys()
    except Exception as ex:
        if dobackup:
            msg = 'The database update failed, rolled back to the original version.'
            ut.printex(ex, msg, iswarning=True)
            ut.remove_file(db_fpath)
            copy_database(db_backup_fpath, db_fpath)
            if clearbackup:
                ut.remove_file(db_backup_fpath)
            raise
        else:
            ut.printex(
                ex,
                ('The database update failed, and no backup was made.'),
                iswarning=False,
            )
            raise
    if dobackup and clearbackup:
        ut.remove_file(db_backup_fpath)


def autogenerate_nth_schema_version(schema_spec, n=-1):
    r"""
    dumps, prints, or diffs autogen schema based on command line

    Args:
        n (int):

    CommandLine:
        python -m wbia.control._sql_helpers --test-autogenerate_nth_schema_version

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control._sql_helpers import *  # NOQA
        >>> from wbia.control import DB_SCHEMA
        >>> # build test data
        >>> schema_spec = DB_SCHEMA
        >>> n = 1
        >>> # execute function
        >>> tablename = autogenerate_nth_schema_version(schema_spec, n)
        >>> # verify results
        >>> result = str(tablename)
        >>> print(result)
    """
    import utool as ut

    logger.info('[_SQL] AUTOGENERATING CURRENT SCHEMA')
    db = get_nth_test_schema_version(schema_spec, n=n)
    # Auto-generate the version skip schema file
    schema_spec_dir, schema_spec_fname = split(schema_spec.__file__)
    schema_spec_fname = splitext(schema_spec_fname)[0]
    # HACK TO GET AUTOGEN COMMAND
    # FIXME: Make this autogen command a bit more sane and not completely
    # coupled with wbia
    autogen_cmd = ut.codeblock(
        """
        python -m wbia.control.{schema_spec_fname} --test-autogen_{funcname} --force-incremental-db-update --write
        python -m wbia.control.{schema_spec_fname} --test-autogen_{funcname} --force-incremental-db-update --diff=1
        python -m wbia.control.{schema_spec_fname} --test-autogen_{funcname} --force-incremental-db-update
        """
    ).format(schema_spec_fname=schema_spec_fname, funcname=schema_spec_fname.lower())
    autogen_text = db.get_schema_current_autogeneration_str(autogen_cmd)

    autogen_fname = '%s_CURRENT.py' % schema_spec_fname
    autogen_fpath = join(schema_spec_dir, autogen_fname)

    dowrite = ut.get_argflag(('-w', '--write', '--dump-autogen-schema'))
    show_diff = ut.get_argflag('--diff')
    num_context_lines = ut.get_argval('--diff', type_=int, default=None)
    show_diff = show_diff or num_context_lines is not None
    dowrite = dowrite and not show_diff

    if dowrite:
        ut.write_to(autogen_fpath, autogen_text)
    else:
        if show_diff:
            if ut.checkpath(autogen_fpath, verbose=True):
                prev_text = ut.read_from(autogen_fpath)
                textdiff = ut.util_str.get_textdiff(
                    prev_text, autogen_text, num_context_lines=num_context_lines
                )
                ut.print_difftext(textdiff)
        else:
            ut.util_print.print_python_code(autogen_text)
        logger.info('\nL___\n...would write to: %s' % autogen_fpath)

    logger.info(' Run with -n=%r to get a specific schema version by index. -1 == latest')
    logger.info(' Run with --write to autogenerate latest schema version')
    logger.info(
        ' Run with --diff or --diff=<numcontextlines> to see the difference between current and requested'
    )
    return db


def get_nth_test_schema_version(schema_spec, n=-1):
    """
    Gets a fresh and empty test version of a schema

    Args:
        schema_spec (module): schema module to get nth version of
        n (int): version index (-1 is the latest)
    """
    from wbia.dtool.sql_control import SQLDatabaseController

    dbname = schema_spec.__name__
    logger.info('[_SQL] getting n=%r-th version of %r' % (n, dbname))
    version_expected = list(schema_spec.VALID_VERSIONS.keys())[n]
    cachedir = ut.ensure_app_resource_dir('wbia_test')
    db_fname = 'test_%s.sqlite3' % dbname
    ut.delete(join(cachedir, db_fname))
    db_uri = 'file://{}'.format(realpath(join(cachedir, db_fname)))
    db = SQLDatabaseController.from_uri(db_uri)
    ensure_correct_version(None, db, version_expected, schema_spec, dobackup=False)
    return db
